#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""cwb.py

definition of the Corpora, Corpus and SubCorpus classes

"""
import logging
import os
from io import StringIO

# requirements
from numpy import maximum, minimum
from pandas import DataFrame, concat, read_csv
from pandas.errors import EmptyDataError

# part of module
from .cache import Cache, generate_idx, generate_library_idx
from .cl import Corpus as Attributes
from .collocates import Collocates
from .concordances import Concordance, format_line
from .counts import Counts, cwb_scan_corpus
from .cqp import start_cqp
from .keywords import Keywords
from .utils import (aggregate_matches, chunk_anchors, correct_anchors,
                    decode, dump_left_join, fold_df, format_roles,
                    intersect_intervals, merge_intervals,
                    preprocess_query, time_it)
from .version import __version__

logger = logging.getLogger(__name__)


class Corpora:
    """Interface to CWB-indexed corpora.

    """

    def __init__(self, cqp_bin='cqp', registry_dir='/usr/local/share/cwb/registry/'):
        """Establish connection to registry and CQP.

        :param str cqp_bin: /path/to/cqp-binary
        :param str registry_dir: /path/to/cwb/registry/

        """

        self.registry_dir = registry_dir
        self.cqp_bin = cqp_bin

    def __str__(self):
        """Method for printing.

        :return: paths, available corpora
        :rtype: str

        """

        corpora = self.show()

        return '\n' + '\n'.join([
            f'registry directory: "{self.registry_dir}"',
            f'cqp binary:         "{self.cqp_bin}"',
            f'found {len(corpora)} corpora:',
            corpora.to_string(),
        ])

    def __repr__(self):
        """Info string

        """
        return self.__str__()

    def show(self):
        """Show all corpora defined in registry and available via CQP
        alongside corpus size (number of tokens).

        :return: available corpora
        :rtype: DataFrame

        """

        # get all corpora defined in registry
        cqp = start_cqp(self.cqp_bin, self.registry_dir)
        corpora = cqp.Exec("show corpora;").split("\n")
        cqp.__del__()

        # check availability and corpus sizes
        sizes = list()
        names = list()
        for corpus_name in corpora:
            try:
                sizes.append(len(Attributes(
                    corpus_name, registry_dir=self.registry_dir
                ).attribute('word', 'p')))
                names.append(corpus_name)
            except SystemError:
                logger.warning(
                    f'corpus "{corpus_name}" defined in registry but not available'
                )

        # create dataframe
        corpora = DataFrame({'corpus': names, 'size': sizes}).set_index('corpus')

        return corpora

    def corpus(self, corpus_name, lib_dir=None, data_dir=None, inval_cache=True):
        """Activate a corpus.

        :param str corpus_name: name of corpus in CWB registry
        :param str lib_dir: /path/to/macros/and/wordlists/
        :param str data_dir: /path/to/data/and/cache/
        :param bool inval_cache: invalidate cache when updating library?

        :return: corpus
        :rtype: Corpus

        """

        return Corpus(corpus_name,
                      lib_dir=lib_dir,
                      cqp_bin=self.cqp_bin,
                      registry_dir=self.registry_dir,
                      data_dir=data_dir,
                      inval_cache=inval_cache)


def init_data_dir(data_dir, corpus_name, lib_dir=None, inval_cache=True):
    """ get a data directory / ensure that the given one complies with schema

    """

    if data_dir is None:
        data_dir = os.path.join("/tmp", "ccc-" + str(__version__))
    if not isinstance(data_dir, str):
        raise ValueError("parameter data_dir must be str")

    # ensure library exists; generate library idx to invalidate cache when updated
    if (not inval_cache) or (lib_dir is None):
        lib_idx = "lib-vanilla"
    else:
        os.makedirs(os.path.join(lib_dir, "wordlists"), exist_ok=True)
        os.makedirs(os.path.join(lib_dir, "macros"), exist_ok=True)
        lib_idx = generate_library_idx(lib_dir, 'lib-', 7)

    # each corpus has its separate directory for each library
    subdir = corpus_name + "-" + lib_idx
    if not data_dir.endswith(subdir):
        data_dir = os.path.join(data_dir, subdir)

    # create directory
    try:
        os.makedirs(data_dir, exist_ok=True)
    except PermissionError:
        logger.error(f'no write permission at "{data_dir}"')
    else:
        return data_dir


class Corpus:
    """Interface to CWB-indexed corpus.

    """

    def __init__(self, corpus_name, lib_dir=None, cqp_bin='cqp',
                 registry_dir='/usr/local/share/cwb/registry/',
                 data_dir=None, inval_cache=True):
        """Establish connection to CQP and corpus attributes, set paths, read
        library.

        Raises KeyError if corpus not in registry.

        data directory contains subdirectories for each corpus and
        each library (/<data-dir>/<corpus>-<lib-idx>/CACHE)
        - marginals_complex
        - dump_from_s_att
        - dump_from_query

        :param str corpus_name: name of corpus in CWB registry
        :param str lib_dir: /path/to/macros/and/wordlists/
        :param str cqp_bin: /path/to/cqp-binary
        :param str registry_dir: /path/to/cwb/registry/
        :param str data_dir: /path/to/data/and/cache/
        :param bool inval_cache: invalidate cache when updating library?

        """

        # preprocess data dir
        data_dir = init_data_dir(data_dir, corpus_name, lib_dir, inval_cache)

        # save paths
        self.data_dir = data_dir
        self.registry_dir = registry_dir
        self.cqp_bin = cqp_bin
        self.lib_dir = lib_dir
        self.inval_cache = inval_cache

        # init (sub-)corpus information
        self.corpus_name = corpus_name
        self.subcorpus_name = None

        # init attributes
        self.attributes = Attributes(self.corpus_name, registry_dir=self.registry_dir)

        # get corpus size
        self.corpus_size = len(self.attributes.attribute('word', 'p'))

        # init cache
        self.cache = Cache(os.path.join(self.data_dir, "CACHE"))

        # init counts
        self.counts = Counts(self.corpus_name, self.registry_dir)

    def __str__(self):
        """Method for printing.

        :return: settings and paths
        :rtype: str

        """

        return '\n' + '\n'.join([
            f'corpus: {self.corpus_name} ({self.corpus_size} tokens)',
            f'data:   {self.data_dir}',
        ])

    def __repr__(self):
        """Info string.

        """
        return self.__str__()

    def size(self):
        """Number of tokens in corpus.

        """
        return self.corpus_size

    def available_macros(self):
        """Get available macros (either system-defined or via library).

        :return: defined macros
        :rtype: list

        """

        cqp = self.start_cqp()
        defined_macros = cqp.Exec("show macro;").split("\n")
        cqp.__del__()

        return defined_macros

    def available_wordlists(self):
        """Get available wordlists.

        :return: defined wordlists
        :rtype: list

        """

        cqp = self.start_cqp()
        defined_wordlists = cqp.Exec("show var;").split("\n")
        cqp.__del__()

        names = sorted(
            [n.rstrip(" =") for n in defined_wordlists if n.startswith("$") and n.endswith(" =")]
        )

        return names

    def available_attributes(self):
        """Get indexed p- and s-attributes. Will be run once when initializing
        the corpus.

        :return: attributes and annotation info
        :rtype: DataFrame

        """

        # use CQP's context descriptor
        cqp = self.start_cqp()
        cqp_ret = cqp.Exec('show cd;')
        cqp.__del__()

        # read as dataframe
        attributes = read_csv(
            StringIO(cqp_ret), sep='\t',
            names=['type', 'attribute', 'annotation', 'active']
        ).fillna(False)

        # post-process Boolean columns
        attributes['active'] = (attributes['active'] == "*")
        attributes['annotation'] = (attributes['annotation'] == '-V')

        return attributes

    def start_cqp(self):
        """Start CQP process.

        :return: CQP process
        :rtype: CQP

        """

        return start_cqp(
            self.cqp_bin,
            self.registry_dir,
            self.data_dir,
            self.corpus_name,
            self.lib_dir,
            self.subcorpus_name
        )

    def copy(self):
        """Get a fresh initialization of the corpus.

        :return: corpus
        :rtype: Corpus

        """

        return Corpus(
            self.corpus_name,
            self.lib_dir,
            self.cqp_bin,
            self.registry_dir,
            self.data_dir,
            self.inval_cache
        )

    ################
    # p-attributes #
    ################
    def cpos2patts(self, cpos, p_atts=['word'], ignore=True):
        """Retrieve p-attributes at corpus position.

        :param int cpos: corpus position to fill
        :param list p_atts: p-attribute(s) to fill position with
        :param bool ignore: whether to return (None, ..) for -1

        :return: p-attribute(s) at cpos
        :rtype: tuple

        """

        return self.counts._cpos2patts(cpos, p_atts, ignore)

    def marginals(self, items=None, p_atts=['word'], flags=0, pattern=False):
        """Extract marginal frequencies. If no items are given, return
        frequencies of all items.

        :param list items: items to get marginals for
        :param str p_att: p-attribute to get frequencies for
        :param int flags: 1 = %c, 2 = %d, 3 = %cd (will activate wildcards)
        :param bool pattern: activate wildcards?

        :return: frequencies of the items in the whole corpus indexed by items
        :rtype: FreqFrame

        """

        # allow lazy evocation
        p_atts = [p_atts] if isinstance(p_atts, str) else p_atts

        # simple
        if len(p_atts) == 1 and items is not None:
            return self._marginals_simple(items, p_atts[0], flags, pattern)

        else:
            return self._marginals_complex(items, p_atts)

    def _marginals_simple(self, items, p_att='word', flags=0, pattern=False):
        """Extract marginal frequencies for given unigrams or unigram patterns
        of a single p-attribute.  0 if not in corpus.  For
        combinations of p-attributes, see marginals_complex.

        :param list items: items to get marginals for
        :param str p_att: p-attribute to get frequencies for
        :param int flags: 1 = %c, 2 = %d, 3 = %cd (will activate wildcards)
        :param bool pattern: activate wildcards?

        :return: frequencies of the items in the whole corpus indexed by items
        :rtype: FreqFrame

        """

        if self.subcorpus_name is not None:
            logger.warning('retrieving corpus marginals, not subcorpus marginals')

        pattern = True if flags > 0 else pattern
        att = self.attributes.attribute(p_att, 'p')

        # loop through items and collect frequencies
        def att_frequency(att, item):
            """ small helper function for list comprehension:
            return frequency of 0 for items that are not in att
            """
            try:
                return att.frequency(item)
            except KeyError:
                return 0

        counts = [
            att_frequency(att, item) if not pattern else
            len(att.find_pattern(item, flags=flags)) for item in items
        ]

        # create dataframe
        df = DataFrame({'freq': counts, p_att: items})
        df = df.set_index(p_att, drop=False)
        df.index.name = 'item'
        df = df.sort_values(['freq', 'item'], ascending=False)

        return df

    def _marginals_complex(self, items, p_atts=['word']):
        """Extract marginal frequencies for p-attribute combinations,
        e.g. ["lemma", "pos"].  0 if not in corpus.  Marginals are
        retrieved using cwb-scan-corpus, result is cached.

        :param list items: list of tuples
        :param list p_atts: list of p-attributes

        :return: counts of the items in the whole corpus indexed by items
        :rtype: FreqFrame

        """

        if self.subcorpus_name is not None:
            logger.warning('retrieving corpus marginals, not subcorpus marginals')

        # get all marginals for p-att combination
        identifier = "-".join(p_atts) + "-marginals"
        df = self.cache.get(identifier)
        if df is not None:
            # get from cache if possible
            logger.info('using cached version of marginals of "%s"' % "_".join(p_atts))
        else:
            # calculate all marginals for p-att combination
            df, R = cwb_scan_corpus(
                self.corpus_name, self.registry_dir, p_atts=p_atts, min_freq=0
            )
            self.cache.set(identifier, df)

        if items is not None:
            # preprocess tuples
            items = [" ".join(i) for i in items] if isinstance(items[0], tuple) else items
            # select relevant rows
            df = df.reindex(items, fill_value=0)

        # sort by frequency
        df = df.sort_values(by=['freq', 'item'], ascending=False)

        return df

    ################
    # s-attributes #
    ################
    def cpos2sid(self, cpos, s_att):
        """Get cwb-id of s-att at cpos. -1 if not present.

        :param int cpos: corpus position
        :param str s_atts: s-attribute to get cwb-id for

        :return: cwb-id
        :rtype: int

        """
        s_attributes = self.attributes.attribute(s_att, "s")
        try:
            return s_attributes.cpos2struc(cpos)
        except KeyError:
            return -1

    ##############
    # subcorpora #
    ##############
    def show_nqr(self):
        """Get subcorpora defined in CQP as DataFrame.

        :return: available subcorpora
        :rtype: DataFrame

        """
        cqp = self.start_cqp()
        cqp_return = cqp.Exec("show named;")

        try:
            df = read_csv(StringIO(cqp_return), sep="\t", header=None)
            df.columns = ["storage", "corpus:subcorpus", "size"]
            crpssbcrps = df["corpus:subcorpus"].str.split(":", n=1).str
            df['corpus'] = crpssbcrps[0]
            df['subcorpus'] = crpssbcrps[1]
            df.drop('corpus:subcorpus', axis=1, inplace=True)
            df = df[['corpus', 'subcorpus', 'size', 'storage']]
        except EmptyDataError:
            logger.info("no subcorpora defined")
            df = DataFrame(columns=['corpus', 'subcorpus', 'size', 'storage'])

        cqp.__del__()

        return df

    ##################
    # CREATING DUMPS #
    ##################
    def dump_from_s_att(self, s_att, annotation=True):
        """Create s-attribute spans as DataFrame of corpus positions.
        Resulting df_dump is indexed by (match, matchend).

        Note that s-attribute values are not indexed by the CWB.
        CWB.CL implementation just iterates over all annotations.

        This method thus creates a dataframe with the
        (non-overlapping) spans encoded as matches

        === (match, matchend) $s_cwbid, $s* ===

        and caches the result.

        $s is only created if annotation is True and attribute is
        actually annotated.

        :param str s_att: s-attribute to get spans and annotation for
        :param bool annotation: whether to retrieve annotation (if present)

        :return: df_dump
        :rtype: DataFrame

        """

        # retrieve from cache if possible
        identifier = s_att + "-spans"
        df = self.cache.get(identifier)
        if df is not None:
            logger.info(f'using cached version of spans of "{s_att}"')
            return df

        # compute
        logger.info(f'creating dataframe of spans of "{s_att}"')

        df = DataFrame(list(self.attributes.attribute(s_att, 's')))
        # two or three columns: 0 (start), 1 (end), 2* (annotation)
        if annotation:
            annotation = (2 in df.columns)  # just to make sure ...
            if not annotation:
                logger.info(f's-att "{s_att}" does not have any annotation')
            else:
                df[2] = df[2].apply(decode)

        # post-process
        df = df.reset_index()
        df = df.rename({'index': s_att + '_cwbid',
                        0: 'match',
                        1: 'matchend',
                        2: s_att}, axis=1)
        df = df.set_index(['match', 'matchend'])

        # put into cache
        self.cache.set(identifier, df)

        return df

    def dump_from_query(self, query, s_query=None, anchors=[],
                        match_strategy='standard', name='Last', save=False,
                        cwb_version=None, propagate_error=False):
        """Execute query, get DataFrame of corpus positions (CWB dump).
        Resulting df_dump is indexed by (match, matchend).

        Note that in the CWB, only two anchors can be active
        simultaneously. The method thus runs the query once with
        anchors set to [0, 1], and then collects the remaining anchor
        points by running the query again on the NQR of the first
        query run for each pair of remaining anchor points.  Optional
        columns for each anchor:

        === (match, matchend), 0*, ..., 9* ===

        The result is cached.

        :param str query: valid CQP query (without 'within' clause)
        :param str s_query: s-attribute used for initial query
        :param list anchors: anchors to search for
        :param str match_strategy: CQP matching strategy
        :param str name: name for NQR
        :param bool save: whether to save NQR to disk

        :return: df_dump
        :rtype: DataFrame

        """

        # we have to give a name to be able to run iteratively when giving several anchors
        name = 'Last' if name is None else name

        # identify query
        if self.subcorpus_name is not None:
            # check subcorpus size to avoid confusion when re-naming
            cqp = self.start_cqp()
            sbcrpssize = cqp.Exec(f"size {self.subcorpus_name}")
            cqp.__del__()
        else:
            sbcrpssize = None

        identifier = generate_idx([
             query, s_query, anchors, match_strategy, self.subcorpus_name, sbcrpssize
        ], prefix="df_dump:")

        # retrieve from cache if possible
        df_dump = self.cache.get(identifier)
        if df_dump is not None:
            logger.info(f'using cached version "{identifier}" of df_dump with {len(df_dump)} matches')
            return df_dump

        # init cqp and set matching strategy
        cqp = self.start_cqp()
        cqp.Exec(f'set MatchingStrategy "{match_strategy}";')

        # get CWB version
        if cwb_version is None:
            cwb_version = {'major': cqp.major_version,
                           'minor': cqp.minor_version,
                           'patch': cqp.beta_version}

        # include optional within clause
        if s_query is None:
            start_query = query
        else:
            start_query = query + ' within ' + s_query

        # get anchors
        remaining_anchors = list(chunk_anchors(anchors, 2))

        # run the query
        logger.info("running CQP query")
        # first run: anchors at 0 and 1 (considering within clause)
        cqp.Exec('set ant 0; ank 1;')
        df_dump = cqp.nqr_from_query(
            query=start_query,
            name=name,
            match_strategy=match_strategy,
            return_dump=True,
            propagate_error=propagate_error
        )
        if propagate_error and isinstance(df_dump, str):
            return df_dump

        logger.info(f"found {len(df_dump)} matches")

        # if there's nothing to return ...
        if len(df_dump) == 0:
            cqp.__del__()
            return df_dump

        df_dump.columns = [0, 1]

        if len(remaining_anchors) > 0:

            # restrict subsequent queries on initial matches
            if (cwb_version['minor'] >= 5) or (cwb_version['minor'] == 4 and cwb_version['patch'] >= 31):
                cqp.Exec(f"Temp = <<{name}/>> ( {query} );")
            elif cwb_version['minor'] == 4 and cwb_version['patch'] >= 16:
                cqp.nqr_activate(self.corpus_name, name)
            else:
                raise NotImplementedError("cannot work with several anchors for CWB versions older than 3.4.16")

            for pair in remaining_anchors:
                logger.info(f".. running query for anchor(s) {str(pair)}")
                # set appropriate anchors
                cqp.Exec(f'set ant {pair[0]};')
                if len(pair) == 2:
                    cqp.Exec(f'set ank {pair[1]};')
                else:
                    cqp.Exec('set ank 1;')

                # dump new anchors
                if (cwb_version['minor'] >= 5) or (cwb_version['minor'] == 4 and cwb_version['patch'] >= 31):
                    cqp.Query(f"Temp = <<{name}/>> ( {query} );")
                elif cwb_version['minor'] == 4 and cwb_version['patch'] >= 16:
                    cqp.Query(f'Temp = <match> ( {query} );')
                else:
                    raise NotImplementedError("cannot work with several anchors for CWB versions older than 3.4.16")
                df = cqp.Dump("Temp")

                # select columns and join to global df
                if len(pair) == 2:
                    df.columns = [pair[0], pair[1]]
                else:
                    df.columns = [pair[0], 1]
                    df = df.drop(1, axis=1)
                df_dump = df_dump.join(df)

            # NA handling
            logger.info("post-processing dataframe")
            # df_dump = df_dump.dropna(axis=1, how='all')
            # it is more reasonable to yield all requested columns
            # (even if some are all NA) -- instead of silently
            # dropping columns
            df_dump = df_dump.fillna(-1)  # , downcast='infer')

        # restrict output to requested anchors
        df_dump = df_dump[anchors]

        # put into cache
        self.cache.set(identifier, df_dump)

        if save:
            cqp.nqr_save(self.corpus_name, name)

        cqp.__del__()

        return df_dump

    #################################################
    # WORKING ON DUMPS ##############################
    #################################################
    def _dump2patt_row(self, row, p_att, start, end):
        """Retrieve p-attribute annotation from start to end of one row.

        :param Series row: dataframe row that contains start and end keys
        :param str p_att: p-attribute to retrieve
        :param str start: key of start column (int or str)
        :param str end: key of end column (int or str)

        :return: p-attribute annotation
        :rtype: str
        """

        # treat missing columns like missing values
        cpos_start = row.get(start, -1)
        cpos_end = row.get(end, -1)

        # if both are missing, return empty string
        if cpos_start == cpos_end == -1:
            return ""
        # if one of them is missing, set start = end or end = start
        if cpos_start == -1:
            cpos_start = cpos_end
        if cpos_end == -1:
            cpos_end = cpos_start

        # lexicalize
        p = self.attributes.attribute(p_att, 'p')
        return " ".join(p[int(cpos_start): int(cpos_end) + 1])

    def dump2patt(self, df_dump, p_att='word', start='match', end='matchend'):
        """Retrieve p-attribute annotation from start to end.

        === (match, matchend), $p ===

        Any additional columns of df_dump are preserved as long as
        there are no conflicts (in which case the original column will
        be overwritten).

        :param DataFrame df_dump: DataFrame with specified columns (possibly as index)
        :param str p_att: p-attribute to retrieve
        :param str start: key of start column (int or str)
        :param str end: key of end column (int or str)

        :return: df_dump
        :rtype: DataFrame

        """

        # pre-process
        index_names = df_dump.index.names
        df = df_dump.reset_index()

        # retrieve attribute
        df[p_att] = df.apply(
            lambda row: self._dump2patt_row(row, p_att, start, end), axis=1
        )

        # post-process
        df = df.set_index(index_names)

        return df

    def dump2satt(self, df_dump, s_att, annotation=True):
        """Retrieve cwb-id, span, and annotation of s-attribute at match.

        Note that this only takes into account the match, not the
        matchend. This is reasonable assuming that the retrieved s-att
        comprises complete matches (match..matchend).

        === (match, matchend), $s_cwbid, $s_span, $s_spanend, $s*  ===

        $s is only created if annotation is True and attribute is
        actually annotated.

        Any additional columns of df_dump are preserved as long as
        there are no conflicts (in which case the original columns
        will be overwritten).

        :param DataFrame df_dump: DataFrame indexed by (match, matchend)
        :param str s_att: s-attribute to retrieve
        :param bool annotation: whether to retrieve annotation of s-att

        :return: df_dump
        :rtype: DataFrame

        """

        # init dataframe
        df = df_dump.reset_index()[['match', 'matchend']]

        # get $s_id
        df[s_att + "_cwbid"] = df.match.apply(
            lambda x: self.cpos2sid(x, s_att)
        )
        nr_missing = (df[s_att + '_cwbid'] == -1).sum()
        logger.info(f's-att "{s_att}" exists at {len(df)-nr_missing} of {len(df)} matches')

        # retrieve where possible
        s_attributes = self.attributes.attribute(s_att, "s")
        df_s = df.loc[~(df[s_att + "_cwbid"] == -1)]
        df = df.join(DataFrame(
            index=df_s.index,
            data=df_s[s_att + "_cwbid"].apply(lambda x: s_attributes[int(x)]).to_list()
        ))

        # two or three columns: 0 (start), 1 (end), 2* (annotation)
        if annotation:
            annotation = (2 in df.columns)  # just to make sure ...
            if not annotation:
                logger.info(f's-att "{s_att}" does not have any annotation')
            else:
                df[2] = df[2].apply(decode)

        # restore original index and post-process
        df = df.set_index(['match', 'matchend'])
        df = df.rename({0: s_att + '_span',
                        1: s_att + '_spanend',
                        2: s_att}, axis=1)

        # join to original dataframe
        df_dump = df_dump.join(df, lsuffix='_bak')
        df_dump = df_dump[[col for col in df_dump if not str(col).endswith('_bak')]]
        if s_att + '_span' in df.columns and s_att + '_spanend' in df.columns:
            df_dump[[
                s_att + '_span', s_att + '_spanend'
            ]] = df[[
                s_att + '_span', s_att + '_spanend'
            ]].fillna(-1)  # , downcast='infer')
        else:
            df_dump[[
                s_att + '_span', s_att + '_spanend'
            ]] = -1

        return df_dump

    def dump2context(self, df_dump, context_left, context_right, context_break):
        """Extend df_dump to context, breaking the context at context_break.

        === (match, matchend), contextid*, context, contextend  ===

        Columns for $context_break, $context_break_cwbid,
        $context_break_span, $context_break_spanend, are also created
        (if applicable).

        Any additional columns of df_dump are preserved.

        Note that in contrast to matches, contexts may overlap.

        For positions where the s-att specified by context_break does
        not exist, context_break is ignored.

        The context creation algorithm does not take into account that
        match and matchend may be part of different spans defined by
        context_break; it only looks at the s-attributes of match, not
        of matchend.

        For the _context_ column (left hand side), the strategy is as
        follows; the strategy for _contextend_ (right hand side) is
        analogous (using context_right, matchend and
        context_break_spanend).

        - if context_break_span is None and context_left is None
          => context = match
        - if context_break_span is None and context_left is not None
          => context = max(0, match - context_left)
        - if context_break_span is not None and context_left is None
          => context = context_break_span
        - if context_break_span is not None and context_left is not None
          => context = max(match - context_left, s_start)

        :param DataFrame df_dump: DataFrame indexed by (match, matchend)
        :param int context_left: maximum context to the left of match
        :param int context_right: maximum context to the right of matchend
        :param str context_break: s-attribute to confine context

        """

        if context_break is None:
            df = df_dump.reset_index()
            # left
            if context_left is None:
                df['context'] = df.match
            else:
                df['context'] = maximum(0, df.match - context_left)
            # right
            if context_right is None:
                df['contextend'] = df.matchend
            else:
                df['contextend'] = minimum(
                    self.corpus_size - 1, df.matchend + context_right
                )
        else:
            # get context confined by s-attribute
            df = self.dump2satt(df_dump, context_break, annotation=False).reset_index()

            # save contextid
            df['contextid'] = df[context_break + '_cwbid']

            # replace -1 to not confuse min()
            df[[
                context_break + '_spanend'
            ]] = df[[context_break + '_spanend']].replace(-1, self.corpus_size + 1)

            # left
            if context_left is None:
                df['context'] = df[context_break + '_span']
            else:
                df['context'] = df.match - context_left
                df['context'] = df[[
                    'context', context_break + '_span'
                ]].max(axis=1)
            # right
            if context_right is None:
                df['contextend'] = df[context_break + '_spanend']
            else:
                df['contextend'] = df.matchend + context_right
                df['contextend'] = df[[
                    'contextend', context_break + '_spanend'
                ]].min(axis=1)

            # revert replacement
            df[[
                context_break + '_spanend'
            ]] = df[[context_break + '_spanend']].replace(self.corpus_size + 1, -1)

        # restore original index
        df = df.set_index(['match', 'matchend'])

        return df

    #################################################
    # QUERY ALIASES #################################
    #################################################
    def query_s_att(self, s_att, values=set(), name=None, overwrite=True):
        """Get s-attribute spans as Dump, optionally restricting the spans by
        matching the provided values against the s-att annotations.

        === (match, matchend), $s_cwbid, $s* ===

        :param str s_att: s-attribute to use for spans
        :param set values: values of s-att annotation to restrict spans to

        :return: dump
        :rtype: Dump

        """

        df_spans = self.dump_from_s_att(s_att)

        # restrict to certain values
        if len(values) > 0:
            if s_att not in df_spans.columns:
                logger.error("cannot restrict spans without annotation")
                df_spans = DataFrame(columns=['match', 'matchend']).set_index(
                    ['match', 'matchend']
                )
            else:
                values = set(values)
                logger.info(f"restricting spans using {len(values)} values")
                df_spans = df_spans.loc[df_spans[s_att].isin(values)]

        # if this is a subcorpus, only return intersection
        if self.subcorpus_name is not None:
            columns_left = list(self.df.columns)
            columns_right = list(df_spans.columns)
            intersection = intersect_intervals(
                self.df.reset_index()[['match', 'matchend'] + columns_left].values.tolist(),
                df_spans.reset_index()[['match', 'matchend'] + columns_right].values.tolist()
            )
            df_spans = DataFrame(intersection)
            df_spans.columns = ['match', 'matchend'] + columns_left + columns_right
            df_spans = df_spans.set_index(['match', 'matchend'])

        # return SubCorpus
        return self.subcorpus(subcorpus_name=name, df_dump=df_spans, overwrite=overwrite)

    def query_cqp(self, cqp_query, context=20, context_left=None,
                  context_right=None, context_break=None, corrections=dict(),
                  match_strategy='standard', name=None, propagate_error=False,
                  overwrite=True):
        """Get query result as (context-extended) Dump (with corrected
        anchors). If a name is given, the resulting NQR (without
        context and before anchor correction) will be written to disk
        in CWB binary format.

        === (match, matchend), 0*, ..., 9*, contextid*, context*, contextend* ===

        :param str query: CQP query
        :param int context: maximum context around match..matchend (symmetric)
        :param int context_left: maximum context left to the match
        :param int context_right: maximum context right to the matchend
        :param str context_break: s-attribute to confine context to
        :param dict corrections: corrections {anchor: offset}
        :param str match_strategy: CQP matching strategy
        :param str name: name for NQR

        :return: dump
        :rtype: Dump

        """
        # preprocess input
        save = False if name is None else True  # save NQR from CQP to disk?
        query_dict = preprocess_query(cqp_query)
        query = query_dict['query']
        s_query = query_dict['s_query']
        anchors = query_dict['anchors']
        s_query = context_break if s_query is None else s_query
        context_left = context if context_left is None else context_left
        context_right = context if context_right is None else context_right

        # get dump from query
        df_dump = self.dump_from_query(
            query=query,
            s_query=s_query,
            anchors=anchors,
            match_strategy=match_strategy,
            name=name,
            save=save,
            propagate_error=propagate_error
        )

        if propagate_error and isinstance(df_dump, str):
            return df_dump

        # empty return?
        if len(df_dump) == 0:
            logger.info("found 0 matches")
            df_dump = DataFrame(columns=['match', 'matchend']).set_index(
                ['match', 'matchend']
            )
        else:
            # extend dump to context
            df_dump = self.dump2context(
                df_dump, context_left, context_right, context_break
            )
            # apply corrections to anchor points
            df_dump = correct_anchors(df_dump, corrections)

        # return SubCorpus
        return self.subcorpus(subcorpus_name=name, df_dump=df_dump, overwrite=overwrite)

    def query(self, cqp_query=None, context=20, context_left=None,
              context_right=None, context_break=None,
              corrections=dict(), s_query=None, s_values=set(),
              match_strategy='standard', name=None, propagate_error=False):
        """Query the corpus, get the result as Dump.

        query wrapper

        - cqp-query (with or without s-query = context-break)

        - s-query (with or without s-values)

        - cqp-query + s-query + s-values (not implemented)

        """
        s_query = context_break if s_query is None else s_query

        if cqp_query is None and s_query is None:
            raise ValueError("you have to provide cqp_query or s_query")

        if cqp_query is None:
            return self.query_s_att(s_query, s_values, name)

        elif len(s_values) == 0:
            return self.query_cqp(cqp_query, context, context_left,
                                  context_right, s_query, corrections,
                                  match_strategy, name, propagate_error)

        else:
            raise NotImplementedError()

    def quick_query(self, s_context, topic_query="", filter_queries=[], match_strategy='longest'):
        """makes sure query result is saved in CWB binary format

        without topic query:
        - finds all s_context spans that contain at least one filter_query
        with topic query:
        - finds all s_context spans that contain topic_query and all filter_queries

        the distinction between with and without topic_query is made
        because in collocation analyses, one topic is combined with
        different other filters; the topic-query then does not have to
        be queried again, only the filters on the topic-context

        :return: identifier (name of NQR on disk)
        :rtype: str

        """

        logger.info("-"*20 + " enter quick-query " + "-"*20)
        if len(topic_query) == 0:

            identifier = generate_idx([self.subcorpus_name, filter_queries, s_context, match_strategy], prefix='Query')

            cqp = self.start_cqp()
            cqp.Exec(f'set MatchingStrategy "{match_strategy}";')
            size = int(cqp.Exec(f'size {identifier};'))

            if size == 0:
                disjunction = " | ".join(['(' + q + ')' for q in filter_queries])
                logger.info(f'disjunction query: {disjunction}')
                cqp.Query(f'{identifier} = {disjunction} within {s_context} expand to {s_context};')
                logger.info(f'.. saving {identifier} in CWB binary format')
                cqp.Exec(f'save {identifier};')

            cqp.__del__()

            return identifier

        # IDENTIFY
        topic_identifier = generate_idx([self.subcorpus_name, topic_query, s_context, match_strategy], prefix='Query')
        filter_identifier = generate_idx([self.subcorpus_name, topic_query, s_context, match_strategy, filter_queries], prefix='Query')

        # CHECK CQP
        cqp = self.start_cqp()
        cqp.Exec(f'set MatchingStrategy "{match_strategy}";')

        # does filter already exist? then we're done
        size = int(cqp.Exec(f'size {filter_identifier};'))
        if size == 0:

            # does topic already exist?
            size = int(cqp.Exec(f'size {topic_identifier};'))
            logger.info(f'topic query: {topic_query}')
            if size == 0:
                # TOPIC
                cqp.Query(f'{topic_identifier} = {topic_query} expand to {s_context};')
                logger.info(f'.. saving {topic_identifier} in CWB binary format')
                cqp.Exec(f'save {topic_identifier};')
            logger.info('.. size: ' + cqp.Exec(f'size {topic_identifier};'))

            # FILTER
            cqp.Exec(f'{filter_identifier} = {topic_identifier};')
            for query in filter_queries:
                logger.info(f'filter query: {query}')
                cqp.Exec(f'{filter_identifier} expand to {s_context};')
                cqp.Query(f'{filter_identifier} = {query} expand to {s_context};')
                logger.info('.. size: ' + cqp.Exec(f'size {filter_identifier};'))

            # SAVE
            logger.info(f'.. saving {filter_identifier} in CWB binary format')
            cqp.Exec(f'save {filter_identifier};')

        cqp.__del__()

        logger.info("-"*20 + " exit quick-query  " + "-"*20)

        return filter_identifier

    @time_it
    def quick_conc(self, topic_query, s_context, window, order=42,
                   cut_off=100, highlight_queries=dict(),
                   filter_queries=dict(), p_show=['word'], s_show=[],
                   match_strategy='longest', htmlify_meta=True, cwb_ids=False):
        """

        :return: concordance lines, each one a dict
        :rtype: list(dict)
        """

        logger.info('quick-conc :: quick concordancing should be quick')

        if order == 'first':
            cqp_order = ""

        elif isinstance(order, int):
            cqp_order = f'randomize {order}'

        else:
            raise ValueError

        if len(topic_query) == 0:

            queries = {**highlight_queries, **filter_queries}

            # INIT CQP
            identifier = self.quick_query(s_context, topic_query="", filter_queries=list(queries.values()), match_strategy=match_strategy)
            cqp = self.start_cqp()

            # init CONTEXT (TextConstellation)
            cqp.Exec(f'sort {identifier} {cqp_order};')
            cqp.Exec(f'cut {identifier} {cut_off};')
            df_context = cqp.Dump(f'{identifier};')
            subcorpus_context = self.subcorpus(None, df_context).set_context(context_break=s_context)
            df_context = subcorpus_context.df[['contextid']]
            df_context = df_context.reset_index().set_index('contextid')

            # HIGHLIGHT
            cqp.Exec(f'{identifier};')
            cqp.Exec(f'set MatchingStrategy "{match_strategy}";')
            for name, query in queries.items():
                cqp.Exec(f'Temp = {query};')
                df_query = cqp.Dump('Temp;')
                if len(df_query) > 0:
                    subcorpus_query = self.subcorpus(None, df_query).set_context(context_break=s_context)
                    df_query = subcorpus_query.df[['contextid']]
                    df_agg = aggregate_matches(df_query, name)
                    df_context = df_context.join(df_agg)
                else:
                    df_context[name] = None
                    df_context[name + '_BOOL'] = False
                    df_context[name + '_COUNTS'] = 0
            cqp.__del__()

            # index by CONTEXT MATCHES
            df = df_context.set_index(['match', 'matchend'])
            names = list(queries.keys())
            names_bool = [n + '_BOOL' for n in names]
            names_count = [n + '_COUNTS' for n in names]
            for b, c in zip(names_bool, names_count):
                df[b] = df[b].fillna(False)
                df[c] = df[c].fillna(0)

            # ACTUAL CONCORDANCING
            conc = Concordance(self.copy(), df)
            lines = conc.lines(form='dict', p_show=p_show, s_show=s_show, order=order, cut_off=cut_off, cwb_ids=cwb_ids)
            output = lines.apply(lambda row: format_roles(row, names, s_show=names_bool+s_show, window=0,
                                                          htmlify_meta=htmlify_meta), axis=1)

        else:

            if len(filter_queries.keys() & highlight_queries.keys()) > 0:
                logger.warning("query names for filter and highlighting overlap")
                highlight_queries = {name: query for (name, query) in highlight_queries.items() if name not in filter_queries.keys()}

            # INIT CQP
            logger.info("quick-conc :: getting context")
            identifier = self.quick_query(s_context, topic_query, list(filter_queries.values()), match_strategy)
            cqp = self.start_cqp()
            cqp.Exec(f'{identifier};')
            cqp.Exec(f'set MatchingStrategy "{match_strategy}";')
            full_size = int(cqp.Exec(f'size {identifier};'))

            if len(filter_queries) == 0:
                cut_off_pre = cut_off
                logger.info(f"quick-conc :: no further filtering according to window, applying cut-off ({cut_off_pre})")

            else:
                # we still have to filter according to window size and highlight all discoursemes
                # we just take a maximum of n*cut_off
                # TODO take next batch if needed
                cut_off_pre = 10 * cut_off
                logger.info(f"quick-conc :: further filtering according to window, applying extended cut-off ({cut_off_pre})")

            cqp.Exec(f'sort {identifier} {cqp_order};')
            cqp.Exec(f'cut {identifier} {cut_off_pre};')
            cqp.Exec(f'{identifier} = {identifier} expand to {s_context};')

            df_context = cqp.Dump(f'{identifier};')
            subcorpus_context = self.subcorpus(df_dump=df_context, overwrite=False).set_context(window, s_context, overwrite=False)
            df_context = subcorpus_context.df[['contextid', 'context', 'contextend']]

            # index by TOPIC MATCHES
            logger.info("quick-conc :: index by topic")
            cqp.Exec(f'Temp = {topic_query};')
            df_query = cqp.Dump('Temp;')
            subcorpus_query = self.subcorpus(df_dump=df_query, overwrite=False).set_context(window, s_context, overwrite=False)
            df_context = dump_left_join(df_context, subcorpus_query.df, 'topic', drop=True, window=window)
            df_context = df_context.set_index(['match_topic', 'matchend_topic'])
            df_context.index.names = ['match', 'matchend']
            df_context = df_context.astype({'offset_topic': 'int'})

            # collect cpos of filter
            logger.info("quick-conc :: collecting cpos of filter")
            matches_filter = dict()
            for name, query in list(filter_queries.items()):
                cqp.Exec(f'Temp = {query};')
                matches_filter[name] = self.subcorpus(df_dump=cqp.Dump('Temp;'), overwrite=False).matches()

            # .. and highlight
            logger.info("quick-conc :: collecting cpos of highlight")
            matches_highlight = dict()
            for name, query in list(highlight_queries.items()):
                cqp.Exec(f'Temp = {query};')
                matches_highlight[name] = self.subcorpus(df_dump=cqp.Dump('Temp;'), overwrite=False).matches()

            logger.info("quick-conc :: formatting")
            output = df_context.apply(
                lambda row: format_line(self, row.name, row, p_show, s_show, matches_filter, matches_highlight, window,
                                        htmlify_meta=htmlify_meta, cwb_ids=cwb_ids),
                axis=1
            )
            output = [line for line in output.values if line is not None]
            output = output[:cut_off]

            actual_size = len(output)
            if cut_off and (cut_off_pre < full_size) and (actual_size < cut_off):
                logger.warning("quick-conc :: potentially missing concordance lines")
                logger.warning(f'- full size:           {full_size}')
                logger.warning(f'- retrieved size       {actual_size}')
                logger.warning(f'- cut-off:             {cut_off}')
                logger.warning(f'- preliminary cut-off: {cut_off_pre}')

        logger.info("quick-conc :: exit")
        return list(output)

    def subcorpus(self, subcorpus_name=None, df_dump=None, overwrite=True):

        return SubCorpus(subcorpus_name, df_dump, self.corpus_name,
                         self.lib_dir, self.cqp_bin, self.registry_dir, self.data_dir, overwrite, self.inval_cache)


class SubCorpus(Corpus):

    def __init__(self, subcorpus_name, df_dump, corpus_name,
                 lib_dir=None, cqp_bin='cqp',
                 registry_dir='/usr/local/share/cwb/registry/',
                 data_dir=None, overwrite=True, inval_cache=True):
        """
        :param str subcorpus_name: name of NQR in CQP
        :param DataFrame df_dump: a "DumpFrame"

        a "DumpFrame" is a dataframe indexed by (match, matchend),
        i.e. non-overlapping corpus spans represented by integers,
        and without missing values.

        CREATION

        dump_from_query:
        - index <int, int> match, matchend: cpos
        - columns <int> 0*, .., 9*: anchor points (optional; missing = -1)

        dump_from_s_att:
        - index <int, int> match, matchend: cpos
        - column <int> $s_cwbid: id
        - column <str> $s: annotation (optional)

        TRANSFORMATION (additional columns are preserved)

        dump2patt:
        - index <int, int> match, matchend: cpos
        - column <str> $p: " "-joined tokens (match..matchend or regions of input columns)

        dump2satt:
        - index <int, int> match, matchend: cpos
        - column <int> $s_cwbid: id (missing = -1) of s_att at match cpos
        - column <int> $s_span: cpos (missing = -1)
        - column <int> $s_spanend: cpos (missing = -1)
        - column <str> $s*: annotation (optional)

        dump2context:
        - index <int, int> match, matchend: cpos
        - column <int> contextid: of id (optional; duplicate of $s_cwbid)
        - column <int> context: cpos
        - column <int> contextend: cpos
        - column <int> $s_cwbid: id (optional; missing = -1)
        - column <int> $s_span: cpos (missing = -1)
        - column <int> $s_spanend: cpos (missing = -1)

        QUERY ALIASES

        query_cqp:
        - index <int, int> match, matchend: cpos
        - columns <int> 0*, .., 9*: anchor points (optional; missing = -1)
        - column <int> contextid: of id (optional)
        - column <int> context: cpos (optional)
        - column <int> contextend: cpos (optional)

        query_s_att:
        - index <int, int> match, matchend: cpos
        - column <int> $s_cwbid: id
        - column <str> $s: annotation (optional)

        """

        super().__init__(corpus_name, lib_dir, cqp_bin, registry_dir, data_dir, inval_cache)

        if (subcorpus_name is None and df_dump is None):
            logger.warning('no subcorpus information provided, returning Corpus')

        else:
            if subcorpus_name is None:  # (and df_dump is not None)
                subcorpus_name = generate_idx([df_dump], prefix='df_')
                self._assign(subcorpus_name, df_dump, overwrite)

            elif df_dump is None:  # (and subcorpus_name is not None)
                cqp = self.start_cqp()
                df_dump = cqp.Dump(subcorpus=subcorpus_name)
                cqp.__del__()

            else:      # (both df_dump and subcorpus_name are given)
                self._assign(subcorpus_name, df_dump, overwrite)

            self.df = df_dump
            self.subcorpus_name = subcorpus_name

    def _assign(self, subcorpus_name, df_dump, overwrite):

        if subcorpus_name in self.show_nqr()['subcorpus'].values:
            # NQR exists
            if overwrite:
                logger.info(f'NQR "{subcorpus_name}" exists, overwriting')
                # create in CQP
                cqp = self.start_cqp()
                cqp.nqr_from_dump(df_dump, subcorpus_name)
                cqp.nqr_save(self.corpus_name, subcorpus_name)
                cqp.__del__()
            # else:
            #     logger.info(f'NQR "{subcorpus_name}" already exists')

        else:
            # create in CQP
            cqp = self.start_cqp()
            cqp.nqr_from_dump(df_dump, subcorpus_name)
            cqp.nqr_save(self.corpus_name, subcorpus_name)
            cqp.__del__()

        if subcorpus_name not in self.show_nqr()['subcorpus'].values:
            logger.error(f'could not assigne NQR "{subcorpus_name}" from dataframe')
        elif overwrite:
            logger.info(f'assigned NQR "{subcorpus_name}" from dataframe')

    def __str__(self):

        return '\n' + '\n'.join([
            f'corpus:    {self.corpus_name} ({self.corpus_size} tokens)',
            f'subcorpus: {self.subcorpus_name} ({self.size()} tokens in {len(self.df)} spans)',
            f'data:      {self.data_dir}'
        ]) + '\n'

    def __repr__(self):
        """Info string.

        """
        return self.__str__()

    def set_context(self, context=None, context_break=None,
                    context_left=None, context_right=None, overwrite=True):
        """Set context in the dump.

        """
        # pre-process context
        if context_left is None:
            context_left = context
        if context_right is None:
            context_right = context

        # set context
        df = self.dump2context(
            self.df, context_left, context_right, context_break
        )

        return self.subcorpus(self.subcorpus_name, df, overwrite=overwrite)

    def set_context_as_matches(self, subcorpus_name=None, overwrite=True):
        """Set match spans in the dump.

        """

        df_context = self.context()

        df_context['match'] = df_context['context']
        df_context['matchend'] = df_context['contextend']
        df_context = df_context.set_index(['match', 'matchend'])

        return self.subcorpus(subcorpus_name=subcorpus_name, df_dump=df_context, overwrite=overwrite)

    def correct_anchors(self, corrections):
        """Correct anchors by integer offsets.

        """
        self.df = correct_anchors(self.df, corrections)

    def breakdown(self, p_atts=['word'], flags="", split=False):
        """Frequency breakdown of match..matchend.

        """

        logger.info('creating frequency breakdown')
        breakdown = self.counts.dump(
            df_dump=self.df,
            start='match', end='matchend',
            p_atts=p_atts, split=split
        )

        breakdown = fold_df(breakdown, flags)

        return breakdown

    def matches(self):
        """
        :return: cpos of (match .. matchend) regions
        :rtype: set
        """

        identifier = generate_idx([self.df.reset_index()[['match', 'matchend']]]) + "-matches"
        f1 = self.cache.get(identifier)
        if not isinstance(f1, set):
            f1 = set()
            for match, matchend in self.df.index:
                f1.update(range(match, matchend + 1))
            self.cache.set(identifier, f1)
        return f1

    def size(self):
        return (self.df.reset_index()['matchend'] - self.df.reset_index()['match'] + 1).sum()

    def context(self):
        """
        :return: cpos of (context .. contextend) regions including matches
        :rtype: DataFrame
        """

        identifier = generate_idx([self.df[['context', 'contextend']]]) + "-contexts"
        df = self.cache.get(identifier)
        if not isinstance(df, DataFrame):
            intervals = self.df[['context', 'contextend']].drop_duplicates().values.tolist()
            df = DataFrame.from_records(merge_intervals(intervals), columns=['context', 'contextend'])
            self.cache.set(identifier, df)
        return df

    def marginals(self, items=None, p_atts=['word'], start='match', end='matchend'):
        """
        :return: counts of (start .. end) regions including matches
        :rtype: DataFrame
        """

        identifier = generate_idx([self.df.reset_index()[[start, end]]]) + "-".join(p_atts) + "-marginals"
        df = self.cache.get(identifier)
        if not isinstance(df, DataFrame):
            df = self.counts.dump(
                self.df, start=start, end=end, p_atts=p_atts, split=True
            )
            self.cache.set(identifier, df)

        if items is not None:
            # preprocess tuples
            items = [" ".join(i) for i in items] if isinstance(items[0], tuple) else items
            # select relevant rows
            df = df.reindex(items, fill_value=0)

        # sort by frequency
        df = df.sort_values(by=['freq', 'item'], ascending=False)

        return df

    def concordance(self, form='simple', p_show=['word'], s_show=[],
                    order='first', cut_off=100, matches=None,
                    slots=None, cwb_ids=False):

        conc = Concordance(
            self.copy(),
            df_dump=self.df
        )

        return conc.lines(
            form=form,
            p_show=p_show,
            s_show=s_show,
            order=order,
            cut_off=cut_off,
            matches=matches,
            slots=slots,
            cwb_ids=cwb_ids
        )

    def collocates(self, p_query=['lemma'], mws=20, window=5,
                   order='O11', cut_off=100, ams=None, min_freq=2,
                   flags=None, marginals='corpus',
                   show_negative=False):

        mws = max(mws, window)

        coll = Collocates(self.copy(), self.df, p_query, mws)

        return coll.show(
            window=window,
            order=order,
            cut_off=cut_off,
            ams=ams,
            min_freq=min_freq,
            flags=flags,
            marginals=marginals,
            show_negative=show_negative
        )

    def keywords(self, p_query=['lemma'], order='O11', cut_off=100,
                 ams=None, min_freq=2, flags=None,
                 marginals='corpus', show_negative=False):

        kw = Keywords(self.copy(), self.df, p_query)

        return kw.show(
            order=order,
            cut_off=cut_off,
            ams=ams,
            min_freq=min_freq,
            flags=flags,
            marginals=marginals,
            show_negative=show_negative
        )


class SubCorpora:
    """
    partitioning of corpus
    - non-overlapping sets of regions (= match..matchend)
    - does not have to be exhaustive → complement
    """
    def __init__(self, corpus_name, mapping,
                 lib_dir=None,
                 cqp_bin='cqp',
                 registry_dir='/usr/local/share/cwb/registry/',
                 data_dir=None,
                 inval_cache=True):
        """
        :param Corpus corpus: CWB corpus
        :param dict mapping: dictionary of {name: query} or DataFrame indexed by (match, matchend) with column 'subcorpus'

        query can be one of
        + <text>-query
        + cqp-query
        """

        self.corpus = Corpus(corpus_name, lib_dir, cqp_bin, registry_dir, data_dir, inval_cache=inval_cache)

        subcorpora = dict()
        if isinstance(mapping, DataFrame):
            column = 'subcorpus'
            for name, df in mapping.groupby(column):
                subcorpora[name] = self.corpus.subcorpus(None, df.drop(column, axis=1))

        elif isinstance(mapping, dict):
            raise NotImplementedError()

        else:
            raise ValueError()

        self.subcorpora = subcorpora

    def keywords(self):
        """
        - one vs. complement
        - one vs. another
        - one vs. rest
        """
        pass

    def query(self, cqp_query=None, context=20, context_left=None,
              context_right=None, context_break=None,
              corrections=dict(), s_query=None, s_values=None,
              match_strategy='longest', propagate_error=False):
        """
        - query each subcorpus defined in partioning
        - note that this finds at most the number of matches as querying the whole corpus
        - because matches might overlap with boundaries
        """

        s_query = context_break if s_query is None else s_query

        if cqp_query is None and s_query is None:
            raise ValueError("you have to provide cqp_query or s_query")

        out = list()
        if cqp_query is None:
            for name, subcorpus in self.subcorpora.items():
                df = self.query_s_att(s_query, s_values).df
                df['subcorpus'] = name
                out.append(df)

        elif s_values is None:
            for name, subcorpus in self.subcorpora.items():
                df = subcorpus.query_cqp(cqp_query, context, context_left,
                                         context_right, s_query, corrections,
                                         match_strategy, propagate_error=propagate_error).df
                df['subcorpus'] = name
                out.append(df)

        else:
            raise NotImplementedError()

        nodes = concat(out)
        # just in case
        nodes['context'] = nodes['context'].astype(int)
        nodes['contextend'] = nodes['contextend'].astype(int)

        self.nodes = nodes

    def concordance(self, form='kwic', p_show=['word'], s_show=[],
                    order='random', cut_off=100, matches=None,
                    slots=None, cwb_ids=False):

        if self.nodes is None:
            raise ValueError("you first have to run a query")

        dfs = list()
        for name, df in self.nodes.groupby('subcorpus'):
            conc = Concordance(
                self.corpus,
                df_dump=df
            ).lines(
                form=form,
                p_show=p_show,
                s_show=s_show,
                order=order,
                cut_off=cut_off,
                matches=matches,
                slots=slots,
                cwb_ids=cwb_ids
            )
            conc['subcorpus'] = name
            dfs.append(conc)

        return concat(dfs)

    def collocates(self, p_query=['lemma'], window=5,
                   order='log_likelihood', cut_off=100, ams=None,
                   min_freq=2, flags=None, marginals='corpus',
                   show_negative=False):
        """
        - find collocates of query matches in each subcorpus
        """

        if self.nodes is None:
            raise ValueError("you first have to run a query")

        dfs = list()
        for name, df in self.nodes.groupby('subcorpus'):
            coll = Collocates(
                self.corpus,
                df,
                p_query=p_query,
                mws=10
            ).show(
                window,
                order,
                cut_off,
                ams,
                min_freq,
                flags,
                marginals,
                show_negative
            )
            coll['subcorpus'] = name
            dfs.append(coll)

        return concat(dfs)



#!/usr/bin/env python3
#cython: language_level=3
#cython: boundscheck=False
#cython: wraparound=False
"""
Taken from docstring for scipy.optimize.cython_optimize module.
"""

from scipy.optimize.cython_optimize cimport brentq

# import math from Cython
from libc cimport math

myargs = {'C0': 1.0, 'C1': 0.7}  # a dictionary of extra arguments
XLO, XHI = 0.5, 1.0  # lower and upper search boundaries
XTOL, RTOL, MITR = 1e-3, 1e-3, 10  # other solver parameters

# user-defined struct for extra parameters
ctypedef struct test_params:
    double C0
    double C1


# user-defined callback
cdef double f(double x, void *args) noexcept:
    cdef test_params *myargs = <test_params *> args
    return myargs.C0 - math.exp(-(x - myargs.C1))


# Cython wrapper function
cdef double brentq_wrapper_example(dict args, double xa, double xb,
                                    double xtol, double rtol, int mitr):
    # Cython automatically casts dictionary to struct
    cdef test_params myargs = args
    return brentq(
        f, xa, xb, <test_params *> &myargs, xtol, rtol, mitr, NULL)


# Python function
def brentq_example(args=myargs, xa=XLO, xb=XHI, xtol=XTOL, rtol=RTOL,
                    mitr=MITR):
    '''Calls Cython wrapper from Python.'''
    return brentq_wrapper_example(args, xa, xb, xtol, rtol, mitr)



from glob import glob

import pandas as pd
import pytest

from ccc import Corpora, Corpus, SubCorpus

from .conftest import DATA_PATH


def get_corpus(corpus_settings, data_dir=DATA_PATH, lib=True):

    if lib:
        lib_dir = corpus_settings.get('lib_dir', None)
    else:
        lib_dir = None

    return Corpus(
        corpus_settings['corpus_name'],
        registry_dir=corpus_settings['registry_dir'],
        lib_dir=lib_dir,
        data_dir=data_dir
    )


###############################################
# INIT ########################################
###############################################

@pytest.mark.init
def test_corpora(germaparl):
    corpora = Corpora(registry_dir=germaparl['registry_dir'])
    assert type(corpora.show()) == pd.DataFrame
    assert "GERMAPARL1386" in corpora.show().index


@pytest.mark.init
def test_corpus(germaparl):
    corpus = get_corpus(germaparl)
    assert corpus.corpus_size > 1000


@pytest.mark.init
def test_get_corpus(germaparl):
    corpus = get_corpus(germaparl)
    assert corpus.corpus_size > 1000


@pytest.mark.init
def test_corpus_descriptor(germaparl):
    corpus = get_corpus(germaparl)
    assert isinstance(corpus.available_attributes(), pd.DataFrame)


@pytest.mark.init
def test_data_dir(germaparl):
    get_corpus(germaparl, lib=False, data_dir=None)
    paths = glob("/tmp/ccc-*")
    assert len(paths) >= 1


#####################################################
# MACROS AND WORDLISTS ##############################
#####################################################
@pytest.mark.init
def test_macros(germaparl):
    corpus = get_corpus(germaparl, lib=False)
    assert '/np(0)' not in corpus.available_macros()
    corpus = get_corpus(germaparl, lib=True)
    assert '/np(0)' in corpus.available_macros()


@pytest.mark.init
def test_wordlists(germaparl):
    corpus = get_corpus(germaparl, lib=True)
    assert "$parties" in corpus.available_wordlists()


#####################################################
# ATTRIBUTES ########################################
#####################################################

@pytest.mark.attributes
def test_cpos2patt(germaparl):
    corpus = get_corpus(germaparl)
    token = corpus.cpos2patts(124345)
    assert isinstance(token, tuple)
    assert token[0] == 'gilt'


@pytest.mark.attributes
def test_cpos2patts(germaparl):
    corpus = get_corpus(germaparl)
    token = corpus.cpos2patts(124345, ['word', 'pos'])
    assert isinstance(token, tuple)
    assert token == ('gilt', 'VVFIN')


#####################################################
# MARGINALS #########################################
#####################################################
@pytest.mark.marginals
def test_marginals_word(germaparl):
    corpus = get_corpus(germaparl)
    df = corpus.marginals(["Merkel", "Seehofer", "gehen"])
    assert df['freq']['gehen'] == 44
    assert len(df) == 3


@pytest.mark.marginals
def test_marginals_lemma(germaparl):
    corpus = get_corpus(germaparl)
    df = corpus.marginals(["Merkel", "Seehofer", "gehen"], p_atts=['lemma'])
    assert df['freq']['gehen'] == 224
    assert len(df) == 3


@pytest.mark.marginals
def test_marginals_pattern(germaparl):
    corpus = get_corpus(germaparl)
    df = corpus.marginals(["Merkel", "Seehofer", "geh.*"], pattern=True)
    assert df['freq']['geh.*'] == 310


@pytest.mark.marginals
def test_marginals_complex(germaparl):
    corpus = get_corpus(germaparl)
    df = corpus._marginals_complex(
        [("gehen", "VVFIN"), ("Seehofer", "NE"), ("Merkel", "NE")],
        ["lemma", "pos"]
    )
    assert len(df) == 3
    assert df['freq']['gehen VVFIN'] == 186


################################################
# CREATING DUMPS ###############################
################################################

@pytest.mark.dump
def test_dump_from_s_att(germaparl):
    corpus = get_corpus(germaparl)
    df = corpus.dump_from_s_att('text_id')
    assert df.iloc[0]['text_id'] == "i13_86_1_1"


@pytest.mark.dump
def test_dump_from_s_att_wo(germaparl):
    corpus = get_corpus(germaparl)
    df = corpus.dump_from_s_att('p')
    assert df.shape[0] == 7332


@pytest.mark.dump
def test_dump_from_query(germaparl):
    corpus = get_corpus(germaparl)
    df_dump = corpus.dump_from_query(
        query=germaparl['query'],
        s_query=germaparl['s_query'],
        match_strategy='standard'
    )
    assert isinstance(df_dump, pd.DataFrame)
    assert df_dump.shape[0] == 30


@pytest.mark.dump
def test_dump_from_query_1(germaparl):
    corpus = get_corpus(germaparl)
    df_dump = corpus.dump_from_query(
        query='[lemma="Horst"] @1[lemma="Seehofer"]',
        anchors=[1],
        match_strategy='longest'
    )
    assert isinstance(df_dump, pd.DataFrame)
    assert len(df_dump) == 11
    assert 1 in df_dump.columns


@pytest.mark.dump
def test_dump_from_query_anchors(germaparl):
    corpus = get_corpus(germaparl)
    df_dump = corpus.dump_from_query(
        query=germaparl['query_anchor'],
        s_query=germaparl['s_query'],
        anchors=germaparl['anchors'],
        match_strategy='standard'
    )
    assert isinstance(df_dump, pd.DataFrame)
    assert df_dump.shape[0] == 30
    assert all(elem in df_dump.columns for elem in germaparl['anchors'])


@pytest.mark.dump
def test_dump_from_query_anchors_implementation(germaparl):

    corpus = get_corpus(germaparl)
    df_dump1 = corpus.dump_from_query(
        query=germaparl['query_anchor'],
        s_query=germaparl['s_query'],
        anchors=germaparl['anchors'],
        match_strategy='standard'
    )

    corpus = get_corpus(germaparl)
    df_dump2 = corpus.dump_from_query(
        query=germaparl['query_anchor'],
        s_query=germaparl['s_query'],
        anchors=germaparl['anchors'],
        match_strategy='standard',
        cwb_version={'minor': 4, 'patch': 16}
    )

    assert df_dump1.equals(df_dump2)


#################################################
# WORKING ON DUMPS ##############################
#################################################

@pytest.mark.dumpp
def test_dump2satt(germaparl):
    corpus = get_corpus(germaparl)
    df_dump = corpus.dump_from_query(
        query=germaparl['query'],
        s_query=germaparl['s_query'],
        match_strategy='standard'
    )
    df_dump['test'] = None
    df = corpus.dump2satt(df_dump, germaparl['s_meta'])
    assert 'test' in df.columns
    assert len(df) == 30
    assert df.iloc[0]['text_id_span'] == 10628


@pytest.mark.dumpp
def test_dump2satt_2(germaparl):
    corpus = get_corpus(germaparl)
    df_dump = corpus.dump_from_query(
        query=germaparl['query'],
        s_query=germaparl['s_query'],
        match_strategy='longest'
    )
    df = corpus.dump2satt(df_dump, 'p')
    assert df.iloc[0]['p_cwbid'] == 456


@pytest.mark.dumpp
def test_dump2context(germaparl):
    corpus = get_corpus(germaparl)
    df_dump = corpus.dump_from_query(
        query=germaparl['query'],
        s_query=germaparl['s_query'],
        match_strategy='standard'
    )
    df_dump = corpus.dump2context(df_dump, 20, 20, 's')
    assert all(
        elem in df_dump.columns for elem in ['context', 'contextid', 'contextend']
    )


@pytest.mark.dumpp
def test_dump2context2(germaparl):
    corpus = get_corpus(germaparl)
    df_dump = corpus.dump_from_query(
        query=germaparl['query'],
        s_query=germaparl['s_query'],
        match_strategy='standard'
    )
    df_dump = corpus.dump2context(df_dump, 20, 20, 'text_id')
    assert all(elem in df_dump.columns for elem in [
        'context', 'contextid', 'contextend', 'text_id_cwbid', 'text_id'
    ])


@pytest.mark.dumpp
def test_dump2context3(germaparl):
    corpus = get_corpus(germaparl)
    df_dump = corpus.dump_from_query(
        query=germaparl['query'],
        s_query=germaparl['s_query'],
        match_strategy='longest'
    )
    df_dump = corpus.dump2context(df_dump, 20, 20, 's')
    assert all(elem in df_dump.columns for elem in [
        'context', 'contextid', 'contextend', 's_cwbid'
    ])


@pytest.mark.dumpp
def test_dump2context4(germaparl):
    corpus = get_corpus(germaparl)
    df_dump = corpus.dump_from_query(
        query=germaparl['query'],
        s_query=germaparl['s_query'],
        match_strategy='longest'
    )
    df_dump = corpus.dump2context(df_dump, None, 5, 'text_id')
    assert all(elem in df_dump.columns for elem in [
        'context', 'contextid', 'contextend', 'text_id_cwbid', 'text_id'
    ])


@pytest.mark.dumpp
def test_dump2patt(germaparl):
    corpus = get_corpus(germaparl)
    df_dump = corpus.dump_from_query(
        query=germaparl['query'],
        s_query=germaparl['s_query'],
        match_strategy='standard'
    )
    # df = corpus.dump2context(df_dump, 20, 20, 'text_id')
    df_dump = corpus.dump2patt(df_dump)
    assert 'word' in df_dump.columns


@pytest.mark.dumpp
def test_dump2patt2(germaparl):
    corpus = get_corpus(germaparl)
    df_dump = corpus.dump_from_query(
        query=germaparl['query'],
        s_query=germaparl['s_query'],
        match_strategy='standard'
    )
    df_dump = corpus.dump2context(df_dump, 20, 20, 'text_id')
    df_dump = corpus.dump2patt(df_dump, start='context', end='contextend')
    assert 'word' in df_dump.columns


#################################################
# QUERY ALIASES #################################
#################################################

@pytest.mark.query
def test_query_fail(germaparl):
    corpus = get_corpus(germaparl)
    df = corpus.query('"tesdsf"').df
    assert isinstance(df, pd.DataFrame)
    assert len(df) == 0


@pytest.mark.query
def test_query_context_1(germaparl):
    corpus = get_corpus(germaparl)
    df = corpus.query(
        cqp_query=germaparl['query_anchor'],
        context=None
    ).df
    assert isinstance(df, pd.DataFrame)
    columns = germaparl['anchors'] + ['context', 'contextend']
    assert all(elem in df.columns for elem in columns)


@pytest.mark.query
def test_query_context_2(germaparl):
    corpus = get_corpus(germaparl)
    df = corpus.query(
        cqp_query=germaparl['query_anchor'],
        context_left=10,
        context=15
    ).df
    assert isinstance(df, pd.DataFrame)
    columns = germaparl['anchors'] + ['context', 'contextend']
    assert all(elem in df.columns for elem in columns)


@pytest.mark.query
def test_query_context_3(germaparl):
    corpus = get_corpus(germaparl)
    df = corpus.query(
        cqp_query=germaparl['query_anchor'],
        context=None,
        context_break='s'
    ).df
    assert isinstance(df, pd.DataFrame)
    columns = germaparl['anchors'] + ['contextid', 'context', 'contextend']
    assert all(elem in df.columns for elem in columns)


@pytest.mark.query
def test_query_context_4(germaparl):
    corpus = get_corpus(germaparl)
    df = corpus.query(
        cqp_query=germaparl['query_anchor'],
        context=10,
        context_break='s'
    ).df
    assert isinstance(df, pd.DataFrame)
    columns = germaparl['anchors'] + ['contextid', 'context', 'contextend']
    assert all(elem in df.columns for elem in columns)


@pytest.mark.query
def test_query_context_5(germaparl):
    corpus = get_corpus(germaparl)
    df = corpus.query(
        cqp_query='[lemma="Horst"]',
        context=10,
        context_break='s'
    ).df
    assert isinstance(df, pd.DataFrame)


@pytest.mark.query
def test_query_lib(germaparl):

    corpus = get_corpus(germaparl)

    nps = corpus.query('/np[]', context=0, match_strategy='longest')

    assert len(nps.df) == 34432
    parties = corpus.query('$parties', context=0)
    assert len(parties.df) == 2140

    dump = corpus.query(
        cqp_query=germaparl['query_lib'],
        context_break=germaparl['s_query'],
        match_strategy='longest'
    )
    assert isinstance(dump.df, pd.DataFrame)
    assert len(dump.df) == 16


def test_query_anchor(germaparl):

    corpus = get_corpus(germaparl)
    query = r'@1[pos="V.*"] @2"\." </s>'
    d = corpus.query(
        query, context_left=3, context_right=None, context_break='s',
        corrections={2: 1}
    )
    assert (d.df[2] == d.df['contextend']).all()


def test_ccc_quick_query(germaparl):

    corpus = get_corpus(germaparl)

    topic_query = '[lemma="die"]'
    s_context = 's'
    filter_queries = {
        'CSU': '[lemma="CSU"]',
        'CDU': '[lemma="CDU"]'
    }

    identifier = corpus.quick_query(
        s_context=s_context,
        topic_query=topic_query,
        filter_queries=filter_queries.values()
    )

    assert identifier in corpus.show_nqr()['subcorpus'].values


def test_ccc_quick_query_2(germaparl):

    corpus = get_corpus(germaparl)

    s_context = 's'
    queries = {
        'Angela': '[lemma="Angela"]',
        'CSU': '[lemma="CSU"]',
        'CDU': '[lemma="CDU"]'
    }

    identifier = corpus.quick_query(
        s_context=s_context,
        filter_queries=queries.values()
    )

    assert identifier in corpus.show_nqr()['subcorpus'].values


#####################################################
# SUBCORPORA ########################################
#####################################################

@pytest.mark.subcorpus
def test_activate_subcorpus(germaparl):

    corpus = get_corpus(germaparl)

    df1 = corpus.dump_from_query(
        "[lemma='Seehofer']"
    )
    corpus.query(
        "[lemma='Horst'] expand to s",
        name='Horst'
    ).df

    horst = corpus.subcorpus('Horst')
    df2 = horst.dump_from_query(
        "[lemma='Seehofer']"
    )
    assert len(df1) > len(df2)


@pytest.mark.subcorpus
def test_deactivate_subcorpus(germaparl):

    corpus = get_corpus(germaparl)
    # query corpus
    df1 = corpus.dump_from_query(
        '[lemma="die"]',
        germaparl['s_query']
    )

    # query subcorpus
    sein = corpus.query(
        cqp_query="[lemma='sein'] expand to s",
        name='Sein'
    )
    df2 = sein.dump_from_query(
        '[lemma="die"]',
        germaparl['s_query']
    )

    # query corpus
    df3 = corpus.dump_from_query(
        '[lemma="die"]',
        germaparl['s_query']
    )

    assert len(df1) == len(df3)
    assert len(df1) > len(df2)


@pytest.mark.subcorpus
def test_create_cached_nqr(germaparl):

    # problem: if a query runs once without having been given a name,
    # the CQP dump (NQR) is not saved to disk
    # if the same query runs again *with a name*, it should be saved to disk

    corpus = get_corpus(germaparl)

    subcorpus = corpus.query('[lemma="jetzt"]')
    assert isinstance(subcorpus, SubCorpus)
    # assert "Jetzt" not in corpus.show_nqr().values

    subcorpus = corpus.query('[lemma="jetzt"]', name='Jetzt')
    assert isinstance(subcorpus, SubCorpus)
    # assert "Jetzt" in corpus.show_nqr().values


@pytest.mark.subcorpus
def test_nqr_from_s_att(germaparl):

    corpus = get_corpus(germaparl)
    corpus.query_s_att("text_party", values={"CDU", "CSU"}, name="Union")
    assert "Union" in corpus.show_nqr().values
    assert isinstance(corpus.subcorpus("Union"), SubCorpus)


@pytest.mark.subcorpus
def test_subcorpus(germaparl):

    corpus = get_corpus(germaparl)
    subcorpus = corpus.query(
        cqp_query='[lemma="und"]',
        context_break=germaparl['s_query'],
        name='Test'
    )

    # init
    assert isinstance(subcorpus, SubCorpus)
    assert len(subcorpus.df) == 2880
    assert all(elem in subcorpus.df.columns for elem in
               ["s_cwbid", "s_span", "s_spanend", "contextid", "context", "contextend"])


@pytest.mark.subcorpus
def test_subcorpus_keywords(germaparl):

    corpus = get_corpus(germaparl)
    subcorpus = corpus.query(
        cqp_query='[lemma="und"]',
        context_break=germaparl['s_query'],
        name='Test'
    )

    # keywords
    assert len(subcorpus.keywords()) == 1
    assert "und" in subcorpus.keywords().index


@pytest.mark.subcorpus
def test_subcorpus_collocates(germaparl):

    corpus = get_corpus(germaparl)
    subcorpus = corpus.query(
        cqp_query='[lemma="und"]',
        context_break=germaparl['s_query'],
        name='Test'
    )
    assert len(subcorpus.collocates(window=1)) == 100
    assert len(subcorpus.collocates(window=10)) == 100


@pytest.mark.subcorpus
def test_subcorpus_context(germaparl):

    corpus = get_corpus(germaparl)
    subcorpus = corpus.query(
        cqp_query='[lemma="und"]',
        context_break=germaparl['s_query'],
        name='Test'
    )
    assert len(subcorpus.collocates(window=10)) == 100
    assert len(subcorpus.collocates(window=20)) == 100
    # collocates
    subcorpus = subcorpus.set_context(20)
    assert len(subcorpus.collocates()) == 100
    assert len(subcorpus.collocates(window=10)) == 100


@pytest.mark.subcorpus
def test_subcorpus_marginals(germaparl):

    corpus = get_corpus(germaparl)
    assert len(corpus.marginals()) == 14034
    subcorpus = corpus.query(
        cqp_query='[lemma="und"]',
        context_break=germaparl['s_query'],
        name='Test'
    )
    assert len(subcorpus.marginals()) == 2


@pytest.mark.subcorpus
def test_subcorpus_query_s_att(germaparl):

    corpus = get_corpus(germaparl)
    black = corpus.query_s_att("text_party", values={"CDU", "CSU"}, name="Union")
    assert "Union" in corpus.show_nqr().values
    assert isinstance(black, SubCorpus)

    interjection = corpus.query_s_att("p_type", values={"interjection"})
    black_interjection = black.query_s_att("p_type", values={"interjection"})
    assert len(black.matches()) > len(interjection.matches()) > len(black_interjection.matches())



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""discoursemes.py

mostly support for MMDA frontend

"""
import logging

# requirements
from association_measures.measures import score
from pandas import DataFrame, concat

# part of module
from . import Corpus, SubCorpus
from .cache import generate_idx
from .collocates import Collocates, dump2cooc
from .concordances import Concordance
from .utils import (aggregate_matches, dump_left_join, format_cqp_query,
                    format_roles, group_lines)

logger = logging.getLogger(__name__)


########################################################
# (1) FOCUS ON TOPIC: DIRECTED (IE INDEXED) DATAFRAMES #
########################################################
# init(dump, name='topic')
# input : dump.df: === (m, me) ci c ce ==
# output: self.df: === (m, me) ci c ce m_topic me_topic o_topic ===
# with duplicate indices:

#   (m   me)     ci        c       ce      m_t   me_t   o_t
#  638  638      18      629      672      638    638     0
#       638      18      629      672      640    640     2
#  640  640      18      629      672      638    638    -2
#       640      18      629      672      640    640     0
# 1202 1202      36     1195     1218     1202   1202     0
# ...   ...     ...      ...      ...      ...    ...   ...

##########################################################
# add_discourseme(dump, name, drop=True)
##########################################################
# input:  === (m, me) ci c ce m_t me_t o_t ==
#         === (m_d, me_d) ci ==
# output: === (m, me) ci c ce m_t me_t o_t m_d me_d o_d ==

#   (m   me)     ci       c      ce    m_t   me_t    o_t     m_d   me_d   o_d
#  638  638      18     629     672    638    638      0     636    636    -2
#       638      18     629     672    640    640      2     636    636    -2
#  640  640      18     629     672    638    638     -2     636    636    -4
#       640      18     629     672    640    640      0     636    636    -4
# 1202 1202      36    1195    1218   1202   1202      0    1205   1205     3
#  ...  ...     ...     ...     ...    ...    ...    ...     ...    ...   ...

##########################################################
# add_discourseme(dump, name, drop=False)
##########################################################
# input:  === (m, me) ci c ce m_t me_t o_t ==
#         === (m_d, me_d) ci ==
# output: === (m, me) ci c ce m_t me_t o_t m_d me_d o_d ==

#     (m     me)     ci        c       ce     m_t    me_t      o_t    m_d    me_d    o_d
#    638    638      18      629      672     638     638        0    636     636     -2
#           638      18      629      672     640     640        2    636     636     -2
#    640    640      18      629      672     638     638       -2    636     636     -4
#           640      18      629      672     640     640        0    636     636     -4
#   1202   1202      36     1195     1218    1202    1202        0   1205    1205      3
# ...       ...     ...      ...      ...     ...     ...       ...   ...     ...    ...
# 149292 149292    7317   149286   149306  149290  149290       -2   <NA>    <NA>   <NA>
#        149292    7317   149286   149306  149292  149292        0   <NA>    <NA>   <NA>

##########################################################
# group_lines()
##########################################################
# === (m, me) ci c ce m0 m0e o0 m1 me1 o1 m2 me2 o2 ===
# with duplicate indices to
# === (m, me) ci c ce d0 d1 d2 ===
# without duplicate indices, where
# d0 = {(o0, m0, m0e), (o0, m0, m0e), ...}
# d1 = {(o1, m1, m1e), ...}
# ...

#####################################################
# (2) OUTER JOIN ON CONTEXTID: TEXTUAL COOCCURRENCE #
#####################################################

#           d_COUNTS                                                  d      d_BOOL
# context_id
# 1061             8  {(33025, 33025), (33027, 33027), (32998, 32998...        True
# 545              4  {(16933, 16933), (16935, 16935), (16919, 16919...        True
# 6815             4  {(132976, 132976), (132969, 132969), (132978, ...        True
# 6548             4  {(123407, 123407), (123394, 123394), (123409, ...        True
# 6986             4  {(139381, 139381), (139383, 139383), (139396, ...        True
# ...            ...                                                ...         ...
# 2259             1                                   {(79339, 79339)}        True
# 1158             1                                   {(36252, 36252)}        True
# 2201             1                                   {(77579, 77579)}        True
# 6636             1                                 {(126185, 126185)}        True
# 1529             1                                   {(51388, 51388)}        True


class Discourseme():

    def __init__(self, cwb_id, name, df):
        """
        :param str name: name of the discourseme
        :param DataFrame df:
        """
        pass

    def breakdown(self, p_atts):
        pass

    def concordance(self, context, p_atts, form):
        pass

    def meta(self, s_atts):
        pass

    def collocates(self, context, form):
        pass


class Constellation_:

    # sqlite database
    cwb_id = ""
    name = ""
    match = 0
    matchend = 1

    def add_discourseme(self, name, df):
        pass

    def breakdown(self, p_atts=['lemma'], flags='', form='dataframe'):
        pass

    def concordance(self, names=None, p_atts=['word'], s_atts=[],
                    s_context=None, context_left=10, context_right=10,
                    order='random', random_seed=42, number=100,
                    form='dataframe'):
        """
        filtered by names (None=all)
        kwic-focus on match, matchend of each name
        other discoursemes are highlighted
        """
        pass

    def meta(self, s_att, pos='match', form='dataframe'):
        """
        dataframe: s_att_value, frequency, ipm
        """

        # # init corpus
        # corpus = Corpus(corpus_name, lib_path, cqp_bin, registry_path, data_path)

        # # init discourseme constellation
        # topic_query = format_cqp_query(topic_items,
        #                                p_query=p_query, s_query=s_query,
        #                                flags=flags_query, escape=escape)
        # dump = corpus.query(topic_query, context=None)
        # # get meta data
        # meta = dump.concordance(
        #     s_show=s_show, form='simple', order=order, cut_off=cut_off
        # )

        # # tabulate
        # output = list()
        # for s in s_show:

        #     # tabulate number of tokens
        #     # TODO: no need to calculate for each text_x, text_y, etc.
        #     lengths = corpus.query_s_att(s)
        #     lengths = lengths.df.reset_index()
        #     lengths['nr_tokens'] = lengths['matchend'] - lengths['match'] + 1

        #     # absolute frequency in each subcorpus
        #     m = meta[s].value_counts()

        #     if len(m) < 100:
        #         m = m.to_frame()
        #         m.columns = ['frequency']
        #         ell = lengths.groupby(s)[['nr_tokens']].agg(sum)
        #         ell = ell.join(m, how='outer')
        #         ell = ell.fillna(0, downcast='infer')
        #         ell['IPM'] = round(ell['frequency'] / ell['nr_tokens'] * 10**6, 2)
        #         ell.index.name = s
        #         ell = ell.reset_index()
        #         output.append(ell.to_html(index=False, bold_rows=False))

        pass

    def collocates(self, window=5, p_atts=['lemma'], flags='',
                   ams=None, order='log_likelihood', cut_off=None, form='dataframe',
                   reference=None):
        """
        if reference is None: reference are marginals
        """
        pass

    def associations(self, reference=None):
        """
        if reference is None: reference are marginals
        """
        pass


class Constellation:

    def __init__(self, dump, name='topic'):
        """
        param Dump dump: dump with dump.corpus, dump.df: == (m, me) ci, c, ce ==
        param str name: name of the node
        """

        self.corpus = dump
        # init with given dump
        self.df = dump.df[['contextid', 'context', 'contextend']].astype("Int64")
        # init added discoursemes
        self.discoursemes = {}
        # the topic is treated as common discourseme
        self.add_discourseme(dump, name=name)

    def __str__(self):
        return (
            "\n" + f"a constellation with {len(self.df)} match indices" + "\n" +
            f"{len(self.discoursemes)} registered discourseme(s)" + "\n" +
            "\n".join([f"- '{d}' with {len(self.discoursemes[d].df)} matches" for d in self.discoursemes])
        )

    def __repr__(self):
        """Info string

        """
        return self.__str__()

    def add_discourseme(self, dump, name='discourseme', drop=True, window=None):
        """
        :param Dump dump: dump.df: == (m, me) ci ==
        :param str name: name of the discourseme
        :param bool drop: remove matches without all discoursemes in node context
        """

        # register discourseme
        if name in self.discoursemes.keys():
            logger.error(f'name "{name}" already taken; cannot register discourseme')
            return

        self.discoursemes[name] = dump
        self.df = dump_left_join(self.df, dump.df, name,
                                 drop=drop, window=window)

    def breakdown(self, p_atts=['word'], flags=""):

        breakdowns = list()
        for idx, dump in self.discoursemes.items():
            d = dump.breakdown(p_atts=p_atts, flags=flags)
            d['discourseme'] = idx
            breakdowns.append(d)
        breakdown = concat(breakdowns)

        return breakdown

    def concordance(self, window=5,
                    p_show=['word', 'lemma'], s_show=[],
                    order='random', cut_off=100, random_seed=42,
                    htmlify_meta=False, cwb_ids=True, form='dict'):
        """Retrieve concordance lines for constellation.

        :param int window: cpos further away from node will be marked 'out_of_window'

        :return: concordance lines
        :rtype: list
        """

        # convert dataframe
        hkeys = list(self.discoursemes.keys())
        df = group_lines(self.df, hkeys)
        conc = Concordance(self.corpus.copy(), df)
        lines = conc.lines(form=form, p_show=p_show, s_show=s_show, order=order, cut_off=cut_off, cwb_ids=cwb_ids)

        if form == 'dict':
            lines = list(lines.apply(lambda row: format_roles(row, hkeys, s_show, window, htmlify_meta=htmlify_meta), axis=1))

        return lines

    def collocates(self, windows=[3, 5, 7], p_show=['lemma'],
                   flags=None, ams=None, min_freq=2,
                   order='log_likelihood', cut_off=None,
                   marginals='corpus'):
        """Retrieve collocates for constellation.

        :param int window: window around node for pre-selected matches

        :return: collocates
        :rtype: list of DataFrames
        """

        # get df_cooc, f1_set, node_freq from cache if possible
        identifier = generate_idx(
            [self.df.reset_index()[['match', 'matchend', 'context', 'contextend']], p_show, min_freq, self.discoursemes.values()]
        )
        f1_set = self.corpus.cache.get(identifier + "-f1_set")
        df_cooc = self.corpus.cache.get(identifier + "-df_cooc")
        node_freq = self.corpus.cache.get(identifier + "-node_freq")

        # create and cache otherwise
        if not isinstance(df_cooc, DataFrame):

            # get relevant contexts
            logger.info('create cooc matrix')
            df_dump = self.df.drop_duplicates(subset=['context', 'contextend'])
            df_cooc, f1_set = dump2cooc(df_dump)

            logger.info('get cpos that are consumed by discoursemes')
            for idx, disc in self.discoursemes.items():
                # print(idx)
                cpos_total = disc.matches()  # cpos of discoursemes in whole corpus (if not approximate)
                # cpos_in = set(df_cooc['cpos']).intersection(cpos_total)  # cpos of discoursemes in topic context (respecting window)
                # breakdown = disc.breakdown(p_atts=p_show)  # breakdown of discourseme
                # print(len(cpos_total))
                # print(len(cpos_in))
                # print(breakdown)
                f1_set.update(cpos_total)
                # print(len(f1_set))
            df_cooc = df_cooc.loc[~df_cooc['cpos'].isin(f1_set)]

            # count node freqs
            node_freq = self.corpus.counts.cpos(f1_set, p_show)

        # determine collocates
        collocates = Collocates(
            corpus=self.corpus.copy(),
            df_dump=None,
            p_query=p_show,
            mws=max(windows),
            df_cooc=df_cooc,
            f1_set=f1_set,
            node_freq=node_freq
        )
        output = dict()
        for window in windows:
            output[window] = collocates.show(
                window=window,
                order=order,
                cut_off=cut_off,
                ams=ams,
                min_freq=min_freq,
                flags=flags,
                marginals=marginals
            )

        return output


class TextConstellation:

    def __init__(self, dump, s_context, name='topic'):
        """
        param Dump dump: dump with dump.corpus, dump.df: == (m, me) ci, c, ce ==
        param str name: name of the node
        """

        if isinstance(dump, dict):
            dumps = dump.copy()
            name = list(dumps.keys())[0]
            dump = dumps.pop(name)

        elif isinstance(dump, SubCorpus):
            dumps = {}

        else:
            raise ValueError()

        # create
        self.s_context = s_context
        self.corpus = dump
        self.N = len(self.corpus.attributes.attribute(s_context, 's'))

        # init discoursemes
        self.discoursemes = {name: dump}

        try:
            self.df = aggregate_matches(dump.df, name)
        except KeyError:        # no matches
            self.df = DataFrame()

        for name, dump in dumps.items():
            self.add_discourseme(dump, name)

    def add_discourseme(self, dump, name='discourseme'):

        # register discourseme
        if name in self.df.columns:
            logger.error(f'name "{name}" already taken; cannot register discourseme')
            return

        try:
            dump = dump.set_context(context_break=self.s_context)
            df = aggregate_matches(dump.df, name)
        except KeyError:        # no matches
            df = DataFrame()

        if df.empty:
            return
        elif self.df.empty:
            self.df = df
        else:
            df = self.df.join(df, how='outer')

        self.discoursemes[name] = dump
        self.df = df.sort_index()

    def breakdown(self, p_atts=['word'], flags=""):

        breakdowns = list()
        for idx, dump in self.discoursemes.items():
            d = dump.breakdown(p_atts=p_atts, flags=flags)
            d['discourseme'] = idx
            breakdowns.append(d)
        breakdown = concat(breakdowns)

        return breakdown

    def concordance(self, window=0, p_show=['word', 'lemma'],
                    s_show=[], order='random', cut_off=100,
                    random_seed=42, htmlify_meta=False):

        # join context..contextend
        contexts = self.corpus.dump_from_s_att(self.s_context, annotation=False)
        contexts.columns = ['contextid']
        contexts = contexts.reset_index().set_index('contextid')
        df = self.df.join(contexts)

        # index by CONTEXT MATCHES
        df = df.set_index(['match', 'matchend'])
        names = list(self.discoursemes.keys())
        names_bool = [n + '_BOOL' for n in names]
        names_count = [n + '_COUNTS' for n in names]
        for b, c in zip(names_bool, names_count):
            df[b] = df[b].fillna(False)
            df[c] = df[c].fillna(0)

        # ACTUAL CONCORDANCING
        conc = Concordance(self.corpus.copy(), df)
        lines = conc.lines(form='dict', p_show=p_show, s_show=s_show, order=order, cut_off=cut_off)
        output = lines.apply(lambda row: format_roles(
            row, names, s_show=names_bool+s_show, window=0, htmlify_meta=htmlify_meta
        ), axis=1)

        return list(output)

    def associations(self, ams=None, min_freq=2,
                     order='log_likelihood', cut_off=None):

        counts = self.df[[c for c in self.df.columns if c.endswith("_BOOL")]]

        if len(counts.columns) == 1:
            logger.error('cannot calculate associations has only one registered discourseme')
            return None

        counts.columns = [c.split("_BOOL")[0] for c in counts.columns]
        cooc = counts.fillna(False)

        # TODO triangular matrix
        tables = list()
        for name in counts.columns:
            # TODO super slow for pandas 2.0, but why?
            table = round(textual_associations(cooc, self.N, name).reset_index(), 2)
            table['node'] = name
            tables.append(table)
        tables = concat(tables)

        tables = tables[
            ['node', 'candidate'] + [d for d in tables.columns if d not in ['node', 'candidate']]
        ]

        return tables


def textual_associations(cooc, N, column):
    """Textual association.

    """
    f1 = cooc[column].sum()
    candidates = [c for c in cooc.columns if c != column]
    records = list()
    for candidate in candidates:
        f2 = cooc[candidate].sum()
        f = (cooc[[column, candidate]].sum(axis=1) == 2).sum()
        records.append({
            'candidate': candidate,
            'f1': f1,
            'f2': f2,
            'f': f,
            'N': N
        })
    contingencies = DataFrame(records).set_index('candidate')
    measures = score(contingencies, freq=True)
    contingencies = contingencies.join(measures.drop(['N'], axis=1))

    return contingencies


def create_constellation(corpus_name,
                         # discoursemes
                         topic_discourseme,
                         filter_discoursemes,
                         additional_discoursemes,
                         # context settings
                         s_context,
                         context=20,
                         window=None,
                         approximate=False,
                         # CWB settings
                         match_strategy='longest',
                         lib_dir=None,
                         cqp_bin='cqp',
                         registry_dir='/usr/local/share/cwb/registry/',
                         data_dir=None,
                         # for creating from items rather than queries
                         querify=True,
                         p_query='word',
                         s_query=None,
                         flags='%cd',
                         escape=True):
    """simple constellation creator. returns a Constellation() if a
    topic_discourseme is given, otherwise a TextConstellation(). Note
    that for TextConstellations, there is no difference between
    additional discoursemes and filter discoursemes (instead, a
    boolean column for each discourseme is added).

    :param dict topic_discourseme: used for init
    :param dict filter_discoursemes: inner join
    :param dict additional_discourseme: left join

    """

    # translate items into queries?
    if querify:

        # copy so local changes don't affect input
        topic_discourseme = topic_discourseme.copy()
        filter_discoursemes = filter_discoursemes.copy()
        additional_discoursemes = additional_discoursemes.copy()

        # pre-process parameters
        s_context = s_query if not s_context else s_context
        s_query = s_context if s_query is None else s_query

        # create queries
        for disc_name, disc_items in topic_discourseme.items():
            disc_query = format_cqp_query(
                disc_items, p_query=p_query, s_query=s_query, flags=flags, escape=escape
            )
            topic_discourseme[disc_name] = disc_query

        for disc_name, disc_items in filter_discoursemes.items():
            disc_query = format_cqp_query(
                disc_items, p_query=p_query, s_query=s_query, flags=flags, escape=escape
            )
            filter_discoursemes[disc_name] = disc_query

        for disc_name, disc_items in additional_discoursemes.items():
            disc_query = format_cqp_query(
                disc_items, p_query=p_query, s_query=s_query, flags=flags, escape=escape
            )
            additional_discoursemes[disc_name] = disc_query

    # init corpus
    corpus = Corpus(corpus_name, lib_dir, cqp_bin, registry_dir, data_dir)

    # topic -> Constellation()
    if len(topic_discourseme) > 0:

        if len(topic_discourseme) > 1:
            raise ValueError("only one topic discourseme can be given")

        # init with topic
        topic_name = list(topic_discourseme.keys())[0]
        topic_query = topic_discourseme[topic_name]
        topic_dump = corpus.query_cqp(
            topic_query,
            context=context,
            context_break=s_context,
            match_strategy=match_strategy
        )
        const = Constellation(topic_dump, topic_name)

        # only count on subcorpus of context of matches
        if approximate:
            sub = topic_dump.df.set_index(['context', 'contextend'])
            corpus = corpus.subcorpus('TempRestriction', df_dump=sub)

        # add filter discoursemes
        for disc_name, disc_query in filter_discoursemes.items():
            disc_dump = corpus.query_cqp(
                disc_query,
                context=None,
                context_break=s_context,
                match_strategy=match_strategy
            )
            if len(disc_dump.df) > 0:
                const.add_discourseme(disc_dump, disc_name, drop=True, window=window)

        # add additional discoursemes
        for disc_name, disc_query in additional_discoursemes.items():
            disc_dump = corpus.query_cqp(
                disc_query,
                context=None,
                context_break=s_context,
                match_strategy=match_strategy
            )
            if len(disc_dump.df) > 0:
                const.add_discourseme(disc_dump, disc_name, drop=False)

    # no topic -> TextConstellation()
    else:

        # no filter implemented: all discoursemes are equal
        discoursemes = {**filter_discoursemes, **additional_discoursemes}

        # init with arbitrary topic
        topic_name = list(discoursemes.keys())[0]
        topic_query = discoursemes.pop(topic_name)
        topic_dump = corpus.query_cqp(
            topic_query,
            context=context,
            context_break=s_context,
            match_strategy=match_strategy
        )
        const = TextConstellation(topic_dump, s_context, topic_name)

        # add further discoursemes
        for disc_name, disc_query in discoursemes.items():
            disc_dump = corpus.query_cqp(
                disc_query,
                context=None,
                context_break=s_context,
                match_strategy=match_strategy
            )
            if len(disc_dump.df) > 0:
                const.add_discourseme(disc_dump, disc_name)

    return const



from tempfile import NamedTemporaryFile

import pandas as pd
import pytest

from ccc.counts import (cwb_lexdecode, cwb_scan_corpus, read_freq_list,
                        score_counts)
from ccc.cwb import Corpus
from ccc.utils import format_cqp_query

from .conftest import DATA_PATH


def get_corpus(corpus_settings, data_dir=DATA_PATH):

    return Corpus(
        corpus_settings['corpus_name'],
        registry_dir=corpus_settings['registry_dir'],
        lib_dir=corpus_settings.get('lib_dir', None),
        data_dir=data_dir
    )


@pytest.mark.cwb_counts
def test_lexdecode_and_read(germaparl):

    df1, R1 = read_freq_list(germaparl['freq_list'],
                             columns=['lemma'])
    df2, R2 = cwb_lexdecode(germaparl['corpus_name'],
                            germaparl['registry_dir'],
                            'lemma')

    assert R1 == R2
    assert df1.equals(df2)


@pytest.mark.cwb_counts
def test_cwb_scan_corpus_subcorpora(germaparl):

    corpus = get_corpus(germaparl)
    cqp = corpus.start_cqp()

    # get a DumpFrame
    cqp.nqr_from_query('[lemma="Helmut"]', name='Tmp')

    # run cwb-scan-corpus on dump
    with NamedTemporaryFile(mode="wt") as f:
        cqp.Exec('dump Tmp > "%s"' % f.name)
        df1, R1 = cwb_scan_corpus(germaparl['corpus_name'],
                                  germaparl['registry_dir'],
                                  f.name)

    # activate a sub-corpus
    cqp.nqr_from_query(query='[lemma="und"] expand to s', name='und')
    cqp.nqr_activate(corpus.corpus_name, 'und')

    # run same query on sub-corpus
    cqp.nqr_from_query('[lemma="Helmut"]', name='Tmp')

    # run cwb-scan-corpus on dump
    with NamedTemporaryFile(mode="wt") as f:
        cqp.Exec('dump Tmp > "%s"' % f.name)
        df2, R2 = cwb_scan_corpus(germaparl['corpus_name'],
                                  germaparl['registry_dir'],
                                  f.name)

    # check that results are different
    assert sum(df2['freq']) != sum(df1['freq'])

    cqp.__del__()


@pytest.mark.marginals
@pytest.mark.cwb_counts
def test_cwb_scan_corpus_marginal(germaparl):

    df1, R1 = cwb_scan_corpus(
        germaparl['corpus_name'],
        germaparl['registry_dir']
    )
    df2, R2 = cwb_scan_corpus(
        germaparl['corpus_name'],
        germaparl['registry_dir'],
        p_atts=['lemma', 'pos']
    )
    assert df1.index.name == 'item'
    assert df2.index.name == 'item'
    assert list(df1.columns) == ['freq', 'word']
    assert list(df2.columns) == ['freq', 'lemma', 'pos']


@pytest.mark.marginals
@pytest.mark.cwb_counts
def test_marginals(germaparl):
    corpus = get_corpus(germaparl)
    freqframe = corpus.marginals(["Helmut", "Kohl", "CDU"])
    assert len(freqframe) == 3
    assert isinstance(freqframe, pd.DataFrame)


@pytest.mark.marginals
@pytest.mark.cwb_counts
def test_marginals_all(germaparl):
    corpus = get_corpus(germaparl)
    freqframe = corpus.marginals(p_atts=['lemma', 'pos'])
    assert isinstance(freqframe, pd.DataFrame)
    assert freqframe.loc['die ART']['freq'] == 11469


@pytest.mark.marginals
@pytest.mark.cwb_counts
def test_marginals_patterns(germaparl):
    corpus = get_corpus(germaparl)
    freqframe = corpus.marginals(["H.*", "Kohl", "CDU"])
    assert len(freqframe) == 3
    assert isinstance(freqframe, pd.DataFrame)
    assert freqframe.loc['H.*']['freq'] == 0
    freqframe = corpus.marginals(["H.*", "Kohl", "CDU"], pattern=True)
    assert len(freqframe) == 3
    assert isinstance(freqframe, pd.DataFrame)
    assert freqframe.loc['H.*']['freq'] == 2466


@pytest.mark.cwb_counts
def test_count_cpos(germaparl):
    corpus = get_corpus(germaparl)
    freqframe = corpus.counts.cpos(list(range(1, 1000)), p_atts=['word'])
    assert isinstance(freqframe, pd.DataFrame)


@pytest.mark.cwb_counts
def test_count_cpos_combo(germaparl):
    corpus = get_corpus(germaparl)
    freqframe = corpus.counts.cpos(list(range(1, 1000)), p_atts=['lemma', 'pos'])
    assert isinstance(freqframe, pd.DataFrame)
    assert list(freqframe.columns) == ['freq'] + ['lemma', 'pos']


@pytest.mark.marginals
@pytest.mark.mwus
@pytest.mark.cwb_counts
def test_count_items(germaparl):

    corpus = get_corpus(germaparl)
    cqp = corpus.start_cqp()

    items = ["Helmut", "Kohl", "CDU"]
    queries = [
        format_cqp_query([item]) for item in items
    ]

    # whole corpus
    counts1 = corpus.marginals(items)
    counts2 = corpus.counts.mwus(cqp, queries)
    assert list(counts1["freq"]) == list(counts2["freq"])

    # subcorpus
    cqp.nqr_from_query(query='[lemma="und"] expand to s', name='und')
    cqp.nqr_activate(corpus.corpus_name, 'und')
    counts1 = corpus.marginals(items)
    counts2 = corpus.counts.mwus(cqp, queries)
    assert counts1.loc[items[0], 'freq'] > counts2.loc[queries[0], 'freq']

    # whole corpus
    cqp.nqr_activate(corpus.corpus_name)
    counts1 = corpus.marginals(items)
    counts2 = corpus.counts.mwus(cqp, queries)
    assert list(counts1["freq"]) == list(counts2["freq"])

    cqp.__del__()


@pytest.mark.mwus
def test_count_mwus_3(germaparl):

    # whole corpus
    corpus = get_corpus(germaparl)
    items = ["Horst Seehofer", r"( CSU )", "CSU", "WES324", "CSU"]
    queries = [
        format_cqp_query([item]) for item in items
    ]

    cqp = corpus.start_cqp()
    counts3 = corpus.counts.mwus(
        cqp,
        queries,
        strategy=3,
        fill_missing=False
    )
    cqp.__del__()

    assert counts3['freq']['CSU'] == 635


@pytest.mark.mwus
@pytest.mark.cwb_counts
def test_count_mwus_strategies(germaparl):

    # whole corpus
    corpus = get_corpus(germaparl)
    items = ["Horst Seehofer", r"( CSU )", "CSU", "WES324"]
    queries = [
        format_cqp_query([item]) for item in items
    ]

    cqp = corpus.start_cqp()
    counts1 = corpus.counts.mwus(
        cqp,
        queries,
        strategy=1,
        fill_missing=False
    )
    assert queries[0] in counts1.index

    counts2 = corpus.counts.mwus(
        cqp,
        queries,
        strategy=2,
        fill_missing=False
    )

    counts3 = corpus.counts.mwus(
        cqp,
        queries,
        strategy=3,
        fill_missing=False
    )

    cqp.__del__()
    assert counts2.equals(counts3)
    assert sum(counts1['freq']) == sum(counts2['freq'])


@pytest.mark.cwb_counts
def test_count_items_subcorpora(germaparl):

    # subcorpus
    corpus = get_corpus(germaparl)
    cqp = corpus.start_cqp()
    dump = corpus.dump_from_s_att("text_role", ["presidency"])
    cqp.nqr_from_dump(dump, 'Presidency')
    cqp.nqr_activate(corpus.corpus_name, 'Presidency')
    items = ["Horst Seehofer", r"( CSU )", "CSU", "WES324", "CSU"]
    queries = [format_cqp_query([item]) for item in items]

    counts1 = corpus.counts.mwus(
        cqp,
        queries,
        strategy=1,
        fill_missing=False
    )
    assert sum(counts1['freq']) > 0

    counts2 = corpus.counts.mwus(
        cqp,
        queries,
        strategy=2,
        fill_missing=False
    )

    counts3 = corpus.counts.mwus(
        cqp,
        queries,
        strategy=3,
        fill_missing=False
    )
    assert counts2.equals(counts3)
    cqp.__del__()


@pytest.mark.cwb_counts
def test_counts_dump_mwu_1_split(germaparl):
    strategy = 1

    corpus = get_corpus(germaparl)
    dump = corpus.dump_from_query('[lemma="die" %cd] [pos="N.*"]')

    df = corpus.counts.dump(dump, p_atts=['word'], split=True, strategy=strategy)
    assert int(df["freq"]["der"]) == 3775

    df = corpus.counts.dump(dump, p_atts=['word', 'lemma'], split=True, strategy=strategy)
    assert int(df["freq"]["der die"]) == 3775


@pytest.mark.cwb_counts
def test_counts_dump_mwu_1_no_split(germaparl):
    strategy = 1

    corpus = get_corpus(germaparl)
    dump = corpus.dump_from_query('[lemma="Helmut"%cd] [lemma="Kohl"%cd]')

    # no split
    df = corpus.counts.dump(dump, p_atts=['word'], split=False, strategy=strategy)
    assert "Helmut Kohl" in df.index

    df = corpus.counts.dump(dump, p_atts=['word', 'pos'], split=False, strategy=strategy)
    assert "Helmut Kohl NE NE" in df.index


@pytest.mark.cwb_counts
def test_counts_dump_mwu_2(germaparl):
    strategy = 2

    corpus = get_corpus(germaparl)
    dump = corpus.dump_from_query('[lemma="Helmut"%cd] [lemma="Kohl"%cd]')

    df = corpus.counts.dump(dump, p_atts=['word'], split=True, strategy=strategy)
    assert df["freq"]["Helmut"] == 6

    df = corpus.counts.dump(dump, p_atts=['word', 'pos'], split=True, strategy=strategy)
    assert df["freq"]["Helmut NE"] == 6

    df = corpus.counts.dump(dump, p_atts=['word'], split=False, strategy=strategy)
    assert "Helmut Kohl" in df.index
    assert df["freq"].iloc[0] == 6

    df = corpus.counts.dump(dump, p_atts=['word', 'pos'], split=False, strategy=strategy)
    assert ("Helmut Kohl NE NE") in df.index
    assert df["freq"].iloc[0] == 6


@pytest.mark.cwb_counts
@pytest.mark.count_matches
def test_counts_matches_mwu_1(germaparl):
    strategy = 1

    corpus = get_corpus(germaparl)
    cqp = corpus.start_cqp()
    cqp.nqr_from_query('[lemma="Helmut"%cd] [lemma="Kohl"%cd]', name='Last')
    cqp.nqr_activate(corpus.corpus_name, 'Last')

    df = corpus.counts.matches(cqp, 'Last', p_atts=['word'], split=False,
                               flags="%cd", strategy=strategy)
    assert "helmut kohl" in df.index
    df = corpus.counts.matches(cqp, 'Last', p_atts=['word'], split=False,
                               strategy=strategy)
    assert "Helmut Kohl" in df.index

    df = corpus.counts.matches(cqp, 'Last', p_atts=['word'], split=True,
                               flags="%cd", strategy=strategy)
    assert "helmut" in df.index

    df = corpus.counts.matches(cqp, 'Last', p_atts=['word'], split=True,
                               strategy=strategy)
    assert "Helmut" in df.index

    cqp.__del__()


@pytest.mark.cwb_counts
@pytest.mark.count_matches
def test_counts_matches_mwu_2(germaparl):
    strategy = 2

    corpus = get_corpus(germaparl)

    cqp = corpus.start_cqp()
    cqp.nqr_from_query('[lemma="Helmut"%cd] [lemma="Kohl"%cd]', name='Last')
    cqp.nqr_activate(corpus.corpus_name, 'Last')
    df = corpus.counts.matches(cqp, 'Last', p_atts=['word'], split=False,
                               flags="%cd", strategy=strategy)
    assert "helmut kohl" in df.index

    df = corpus.counts.matches(cqp, 'Last', p_atts=['word'], split=False,
                               strategy=strategy)
    assert "Helmut Kohl" in df.index

    df = corpus.counts.matches(cqp, 'Last', p_atts=['word'], split=True,
                               flags="%cd", strategy=strategy)
    assert "helmut" in df.index

    df = corpus.counts.matches(cqp, 'Last', p_atts=['word'], split=True,
                               strategy=strategy)
    assert "Helmut" in df.index
    cqp.__del__()


@pytest.mark.cwb_counts
@pytest.mark.count_matches
def test_counts_matches_mwu_3(germaparl):
    strategy = 3

    corpus = get_corpus(germaparl)
    cqp = corpus.start_cqp()
    cqp.nqr_from_query('[lemma="Helmut"%cd] [lemma="Kohl"%cd]', name='Last')
    df = corpus.counts.matches(cqp, 'Last', p_atts=['word'], split=True,
                               flags="%cd", strategy=strategy)
    assert "helmut" in df.index

    df = corpus.counts.matches(cqp, 'Last', p_atts=['word'], split=True,
                               strategy=strategy)
    assert "Helmut" in df.index

    df = corpus.counts.matches(cqp, 'Last', p_atts=['word', 'pos'], split=True,
                               strategy=strategy)
    assert "Helmut NE" in df.index
    cqp.__del__()


@pytest.mark.cwb_counts
def test_counts_mwus(germaparl):
    corpus = get_corpus(germaparl)
    cqp = corpus.start_cqp()
    queries = ['[lemma="Helmut"%cd & pos="NE"] [lemma="Kohl"]', '[lemma="Horst"]']

    df = corpus.counts.mwus(cqp,
                            queries,
                            strategy=1)
    assert df['freq'][queries[1]] == 55

    df = corpus.counts.mwus(cqp,
                            queries,
                            strategy=3,
                            p_atts=['lemma', 'pos'])
    assert df['freq']['Horst NE'] == 55

    df = corpus.counts.mwus(cqp,
                            queries,
                            strategy=3,
                            p_atts=['lemma'])

    assert df['freq']['Horst'] == 55

    cqp.__del__()


@pytest.mark.cwb_counts
def test_cwb_counts(germaparl):
    corpus = get_corpus(germaparl)
    cqp = corpus.start_cqp()
    queries = ['[lemma="Helmut"%cd & pos="NE"] [lemma="Kohl"]', '[lemma="Horst"]']
    df = corpus.counts.mwus(cqp,
                            queries)

    assert df['freq'][queries[1]] == 55
    cqp.__del__()


def test_score_counts(germaparl, empirist):

    df1, R1 = read_freq_list(germaparl['freq_list'])
    df2, R2 = read_freq_list(empirist['freq_list'])
    df = df1[['freq']].rename(columns={'freq': 'f1'}).join(
        df2[['freq']].rename(columns={'freq': 'f2'})
    )
    df['N1'] = R1
    df['N2'] = R2
    df = df.fillna(0)  # , downcast='infer')

    kw = score_counts(df, cut_off=None)
    assert kw['log_likelihood']['die'] == 4087.276827



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""dumps.py

definition of Dump and Dumps classes

TODO: re-write as subcorpora

"""
import logging

from pandas import DataFrame, concat

from .collocates import Collocates

logger = logging.getLogger(__name__)


class Dumps:
    """ collection of dumps, i.e. query matches in different subcorpora """

    def __init__(self, corpus, s_dict=None, s_att='text_id', df_dumps=None):
        """
        :param Corpus corpus: corpus to work on
        :param dict s_dict: dicitonary of {subcorpus_name: set of values for s_att}
        :param str s_att: s-attribute that is used for definition of subcorpora
        """

        self.corpus = corpus.copy()
        self.s_dict = s_dict
        self.s_att = s_att

        dumps = dict()

        if df_dumps is not None:
            logger.info("saving subcorpora ...")
            for i, s in enumerate(df_dumps.keys()):
                logger.info(f"... subcorpus {i+1} of {len(df_dumps)}: %s")
                df_dump = df_dumps[s]
                dumps[s] = corpus.subcorpus(None, df_dump)

        else:
            logger.info("creating subcorpora ...")
            df_spans = corpus.dump_from_s_att(s_att)
            for i, s in enumerate(self.s_dict.keys()):
                logger.info(f"... subcorpus {i+1} of {len(s_dict)}: %s")
                df_dump = df_spans.loc[df_spans[s_att].isin(s_dict[s])]
                dumps[s] = corpus.subcorpus(None, df_dump)

        self.dumps = dumps

    def keywords(self, p_query=['lemma'], order='log_likelihood',
                 cut_off=100, ams=None, min_freq=2, flags=None,
                 subset=None, marginals='corpus', show_negative=False):
        """
        :return: dictionary of {subcorpus_name: table}
        :rtype: dict
        """
        # TODO: multiproc

        subset = list(self.dumps.keys()) if subset is None else subset

        logger.info("computing keyword tables ...")
        tables = list()
        for i, s in enumerate(subset):
            logger.info(f"... table {i+1} of {len(subset)}")
            dump = self.dumps[s]
            table = dump.keywords(
                p_query=p_query, order=order, cut_off=cut_off,
                ams=ams, min_freq=min_freq, flags=flags, marginals=marginals,
                show_negative=show_negative
            )
            if len(table) > 0:
                table['subcorpus'] = s
                tables.append(table.reset_index())

        df = concat(tables).reset_index(drop=True).set_index(['subcorpus', 'item'])

        return df

    def collocates(self, cqp_query, window=5, p_query=['lemma'],
                   order='log_likelihood', cut_off=100, ams=None,
                   min_freq=2, flags=None, subset=None,
                   context_break=None, marginals='corpus',
                   show_negative=False):
        """
        TODO: context / context_left / context_right
        TODO: allow context = None

        reference:
        .. local: window freq. compared to marginals of subcorpus (excl. nodes)
        .. global: window freq. compared to marginals in whole corpus (excl. nodes)
        .. DataFrame:

        :param str reference: 'local' | 'global' | DataFrame
        :return: dictionary of {subcorpus_name: table}
        :rtype: dict
        """
        # TODO: multiproc

        subset = list(self.dumps.keys()) if subset is None else subset
        context_break = self.s_att if context_break is None else context_break

        # run query once and extend dump
        dump_glob = self.corpus.query_cqp(
            cqp_query,
            context=window,
            context_break=context_break
        ).df
        df_glob = self.corpus.dump2satt(dump_glob, self.s_att)

        logger.info("computing collocate tables ...")
        tables = list()
        for i, s in enumerate(subset):
            logger.info(f"... table {i+1} of {len(subset)}")

            # determine reference frequencies
            if isinstance(marginals, str):
                if marginals == 'local':
                    # get local marginals
                    # THIS is expensive, use subcorpus.marginals instead!
                    marginals = self.corpus.counts.dump(
                        self.dumps[s].df, split=True, p_atts=p_query
                    )
                    # marginals = self.corpus.marginals(p_atts=p_query)
                elif marginals == 'corpus':
                    pass
                else:
                    raise NotImplementedError()
            elif not isinstance(marginals, DataFrame):
                raise NotImplementedError()

            # create collocates table
            df_loc = df_glob.loc[
                df_glob[self.s_att].isin(self.s_dict[s])
            ]
            collocates = Collocates(
                self.corpus, df_loc, p_query=p_query, mws=window
            )
            table = collocates.show(
                window=window, order=order, cut_off=cut_off,
                ams=ams, min_freq=min_freq, flags=flags, marginals=marginals
            )
            if len(table) > 0:
                table['subcorpus'] = s
                tables.append(table.reset_index())

        df = concat(tables).reset_index(drop=True).set_index(['subcorpus', 'item'])

        return df



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""counts.py

methods for counting and scoring items.

"""
import logging
import subprocess
from collections import Counter
from io import StringIO
from tempfile import NamedTemporaryFile

# requirements
from association_measures import measures
from pandas import DataFrame, MultiIndex, read_csv

# part of module
from .cl import Corpus as Crps
from .utils import fold_df, time_it

logger = logging.getLogger(__name__)


def count_items(items, names, tuples=True):
    """Get type frequency table of items.

    :param list items: list of values or tuples
    :param list names: name(s) for the attributes to count
    :param bool tuples: treat each item as a tuple?

    :return: frequency counts of items
    :rtype: FreqFrame
    """

    logger.info(f"counting {len(items)} items")
    counts = Counter(items)
    df_counts = DataFrame.from_dict(counts, orient='index', columns=['freq'])

    # transform index
    if tuples:
        df_counts.index = MultiIndex.from_tuples(df_counts.index, names=names)
        df_counts = df_counts.reset_index()
        df_counts['item'] = df_counts[names].agg(' '.join, axis=1)
        df_counts = df_counts.set_index('item')
    else:
        df_counts = df_counts.reset_index()
        df_counts.columns = names + ['freq']
        df_counts = df_counts.set_index(names[0], drop=False)
        df_counts.index.name = 'item'

    df_counts = df_counts[['freq'] + names]

    df_counts = df_counts.sort_values(by=['freq', 'item'], ascending=False)

    return df_counts


def read_freq_list(path, min_freq=2, columns=None):
    """Read frequency list, e.g. output of lexdecode (see below)

    :param str path: path to read from
    :param int min_freq: drop everything that doesn't appear at least this often

    :return: frequency list (index: item, columns: freq) and original size
    :rtype: tuple(DataFrame, int)

    """

    # read data
    logger.info('reading frequency list ...')
    df = read_csv(path, sep="\t", header=None, quoting=3,
                  keep_default_na=False)
    logger.info(f'reading frequency list ... {df.shape[0]} items')

    # get corpus size
    R = df[0].sum()

    # apply frequency threshold
    logger.info('applying frequency threshold ...')
    df = df.loc[df[0] >= min_freq]
    logger.info(f'applying frequency threshold ... {df.shape[0]} items')

    # indexing
    logger.info('combining relevant columns ...')
    col = list(df.columns[1:]) if columns is None else columns
    df.columns = ['freq'] + col
    if len(col) > 1:
        df['item'] = df[col].agg(' '.join, axis=1)
    else:
        df['item'] = df[col]
    df = df.set_index('item')
    logger.info(
        'combining relevant columns ... item="%s"' % " ".join([str(c) for c in col])
    )

    df = df.sort_values(['freq', 'item'], ascending=False)

    return df, R


def cwb_lexdecode(corpus_name, registry_dir,
                  p_att='word', cmd='cwb-lexdecode', min_freq=2):
    """Run cwb-lexdecode: create frequency list of p-attribute.
    CLI: cwb-lexdecode -f -s -P lemma CORPUS_NAME

    :param str corpus_name:
    :param str registry_dir:
    :param str p_att:
    :param str cmd: path to binary
    :param int min_freq:

    :return: frequency list of p-attribute values and corpus size
    :rtype: tuple(DataFrame, int)

    """

    logger.info("running cwb-lexdecode ...")
    command = [cmd, '-f', '-P', p_att, '-r', registry_dir, corpus_name]
    lexdecode = subprocess.Popen(
        command, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    ret = lexdecode.communicate()[0].decode()
    df_counts, R = read_freq_list(StringIO(ret), min_freq=min_freq, columns=[p_att])

    return df_counts, R


def cwb_scan_corpus(corpus_name, registry_dir, path=None,
                    p_atts=['word'], cmd='cwb-scan-corpus', min_freq=2):
    """Run cwb-scan-corpus: create frequency list of p-attribute(s) at corpus positions.
    CLI: cwb-scan-corpus [-R path] CORPUS_NAME p_atts

    Ranges must be sorted and non-overlapping.

    :return: counts of the p-attribute values (or their combinations)
             at the given positions (if any)
    :param str corpus_name:
    :param str registry_dir:
    :param str p_att:
    :param str cmd: path to binary
    :param int min_freq:

    :return: frequency list of p-attribute values and corpus size
    :rtype: tuple(DataFrame, int)

    """

    logger.info("running cwb-scan-corpus ...")
    command = [cmd, '-r', registry_dir]
    if path is not None:
        command += ['-R', path]
    command += [corpus_name] + p_atts
    scan = subprocess.Popen(
        command, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    ret = scan.communicate()[0].decode()
    df_counts, R = read_freq_list(StringIO(ret), min_freq=min_freq, columns=p_atts)

    return df_counts, R


class Counts:
    """All methods return a FreqFrame

    == (item)* freq, p_atts[0], ...  ==

    where item = " ".join(p_atts)

    If index is not split, MWUs are " "-joined.

    TODO: counting with group?

    """
    def __init__(self, corpus_name,
                 registry_dir='/usr/local/share/cwb/registry/'):

        self.corpus_name = corpus_name
        self.registry_dir = registry_dir
        self.attributes = Crps(
            self.corpus_name,
            registry_dir=registry_dir
        )

    def _cpos2patts(self, cpos, p_atts=['word'], ignore=True):
        """Retrieve p-attributes of corpus position.

        :param int cpos: corpus position to fill
        :param list p_atts: p-attribute(s) to fill position with
        :param bool ignore: whether to return (None, .*) for -1

        :return: p-attribute(s) at cpos
        :rtype: tuple

        """

        if cpos == -1 and ignore:
            token = [None] * len(p_atts)
        else:
            token = [
                self.attributes.attribute(p_att, 'p')[cpos] for p_att in p_atts
            ]

        return tuple(token)

    def cpos(self, cpos_list, p_atts=['word']):
        """Create a frequency table for the p-attribute values of the
        cpos-list.

        - strategy: split   /YES; flags  ; combo x

        :param list cpos_list: corpus positions to fill
        :param list p_atts: p-attribute (combinations) to count

        :return: counts of the p_attribute (combinations) of the positions
        :rtype: FreqFrame

        """
        items = [self._cpos2patts(p, p_atts=p_atts) for p in cpos_list]
        df_counts = count_items(items, p_atts)
        return df_counts

    @time_it
    def dump(self, df_dump, start='match', end='matchend',
             p_atts=['word'], split=False, strategy=2):
        """Count tokens in [start .. end] (columns or index columns in df_dump).

        - strategy 1: split NO/YES; flags  ; combo x
        - strategy 2: split   /YES; flags  ; combo x

        :param DataFrame df_dump: corpus positions to fill
        :param str start: column name where to start counting
        :param str end: column name where to end counting
        :param list p_atts: p-attribute (combinations) to count
        :param bool split: token-based count? (default: MWU)
        :param int strategy: strategy 2 (cwb-scan-corpus) is faster,
                             does not support MWU counts though

        :return: counts of the p_att (combin.) in the spans of two columns of the dump
        :rtype: FreqFrame

        """

        # choose strategy
        if strategy == 2 and not split:
            logger.info("dump: cannot use cwb-scan-corpus for MWUs, extracting and splitting tokens instead")
            strategy = 1
        logger.info(f"dump: strategy {strategy}")

        # for working with match, matchend
        df_dump = df_dump.reset_index()

        if strategy == 1:

            logger.info("... extracting tokens")
            ls = df_dump.apply(
                lambda x: [self._cpos2patts(cpos, p_atts) for cpos in range(
                    x[start], x[end] + 1
                )], axis=1
            ).values            # list of list of tuples (p_att_1, p_att_2, ...)

            logger.info("... splitting")
            if split:
                tokens = [token for tokens in ls for token in tokens]
            else:
                tokens = [
                    tuple([" ".join(m) for m in zip(*mwu_list)]) for mwu_list in ls
                ]

            df_counts = count_items(tokens, p_atts)

        elif strategy == 2:
            df_dump = df_dump.reset_index()
            with NamedTemporaryFile(mode="wt") as f:
                logger.info("... writing dump temporarily to disk")
                df_dump[[start, end]].to_csv(f.name, sep="\t", header=None, index=False)
                df_counts, R = cwb_scan_corpus(
                    self.corpus_name, self.registry_dir, f.name, p_atts, min_freq=1
                )

        df_counts = df_counts.sort_values(by=['freq', 'item'], ascending=False)

        return df_counts

    @time_it
    def matches(self, cqp, name, p_atts=["word"], split=False, flags=None, strategy=3):
        """Counts tokens in [match .. matchend] of named subcorpus defined in
        running cqp.

        - strategy 1: split NO/   ; flags x; combo
        - strategy 2: split NO/YES; flags x; combo
        - strategy 3: split   /YES; flags  ; combo x

        :param CQP cqp: running cqp process
        :param list name: name of the subcorpus
        :param list p_atts: p-attribute(-combinations) to count
        :param bool split: token-based count? (default: MWU)
        :param str flags: %c, %d, %cd

        :return: counts of the p_attribute (combinations) of the matches
        :rtype: FreqFrame

        """

        # choose strategy
        combo = len(p_atts) > 1

        #    s f c
        # 1: - - -
        # 1: - x -
        # 2: - - -
        # 2: - x -
        # 2: x - -
        # 2: x x -
        # 3: x - -
        # 3: x - x

        # implemented:
        #    - - - 1,2
        #    - x - 1,2
        #    x - - 2,3
        #    x x - 2
        #    x - x 3

        # not implemented:
        #    - - x
        #    - x x
        #    x x x

        if combo:
            if flags or (not flags and not split):
                raise NotImplementedError(
                    "matches does not support parameter combination:",
                    str(" ".join(['x' if x else '-' for x in [
                        split, len(flags) > 0, combo
                        ]]))
                )

        if strategy == 1:
            if split or combo:
                logger.warning(
                    "matches: cannot use cqp-count"
                )
                strategy = 2
        if strategy == 2:
            if combo:
                logger.warning(
                    "matches: cannot use cqp-tabulate"
                )
                strategy = 3
        if strategy == 3:
            if flags or not split:
                logger.warning(
                    "matches: cannot use cwb-scan-corpus"
                )
                strategy = 2
        logger.info(f"matches: strategy {strategy}")

        if strategy == 1:
            # split NO; flags NO/YES; combo NO
            # generally slow
            logger.info("... cqp is counting")
            cqp_return = cqp.Exec(f'count {name} by {p_atts[0]} {flags};')
            df_counts = read_csv(
                StringIO(cqp_return), sep="\t", header=None,
                names=["freq", "unknown", p_atts[0]]
            )
            df_counts = df_counts.set_index(p_atts[0], drop=False)
            df_counts.index.name = 'item'
            df_counts = df_counts.sort_values(['freq', 'item'], ascending=False)

        elif strategy == 2:
            # split NO/YES; flags NO/YES; combo NO
            # generally faster
            logger.info("... cqp is tabulating")
            cqp_return = cqp.Exec(f'tabulate {name} match .. matchend {p_atts[0]} {flags};')
            logger.info("... splitting tokens")
            if split:           # split strings into tokens
                cqp_return = cqp_return.replace(" ", "\n")
            tokens = cqp_return.split("\n")
            df_counts = count_items(tokens, names=[p_atts[0]], tuples=False)

        elif strategy == 3:
            # split YES; flags NO; combo YES
            # generally fastest
            with NamedTemporaryFile(mode="wt") as f:
                logger.info("... writing dump temporarily to disk")
                cqp.Exec(f'dump {name} > "{f.name}";')
                df_counts, R = cwb_scan_corpus(
                    self.corpus_name, self.registry_dir, f.name, p_atts, min_freq=1
                )

        return df_counts

    @time_it
    def mwus(self, cqp, queries, p_atts=None, fill_missing=True, strategy=1):
        """Calculates frequencies for MWU queries in activated subcorpus.
        queries are a list of valid CQP queries, e.g.

        '[lemma="Angela"%cd & pos="NE"] [lemma="Merkel"%cd & pos="NE"]?'

        - strategy 1: split NO| - ; flags x; combo x; mwu NO
        - strategy 2: split NO| - ; flags x; combo x; mwu YES
        - strategy 3: split NO| - ; flags x; combo  ; mwu YES

        caveat: strategy 1 does not yield breakdown in attributes
        this implies also different indexing

        Strategies:
        - Strategy 1: for each item
          1. run query
          2. get size of NQR via CQP
        - Strategy 2:
          1. run query for all items at the same time
          2. dump df
          3. count_dump()
        - Strategy 3:
          1. run query for all items at the same time
          2. count_matches()

        :param CQP cqp: running cqp process
        :param set queries: set of query strings to get frequency breakdown for
        :param bool fill_missing: count 0 for missing items?
        :param int strategy: strategy to use (see below)

        :return: counts of the queries (strategy 1) or the items in the queries
        :rtype: FreqFrame

        """

        # subcorpus name to use
        name = 'Tmp'

        # choose strategy
        if strategy == 1:
            if p_atts:
                logger.warning(
                    "mwus: cannot get frequency breakdown when not inspecting dump"
                )
                strategy = 2

        # necessary for strategies 2 & 3:
        p_atts = ['word'] if not p_atts else p_atts

        if strategy == 3 and len(p_atts) > 1:
            logger.warning(
                "mwus: cannot combine query when looking at several p-attributes"
            )
            strategy = 2
        logger.info(f"mwus: strategy {strategy}")

        # count
        if strategy == 1:
            logger.info("... running each query")
            freqs = list()
            for query in queries:
                cqp.Exec(f'{name}={query};')
                freq = cqp.Exec(f'size {name};')
                freqs.append(freq)
            df = DataFrame(data=freqs, index=queries, columns=['freq'])
            df.index.name = 'item'
            df['freq'] = df['freq'].astype(int)
            df = df.sort_values(by=['freq', 'item'], ascending=False)

        elif strategy == 2:
            query = "|".join(queries)
            cqp.Exec(f'{name}={query};')
            df_dump = cqp.Dump(name)
            df = self.dump(df_dump, start='match', end='matchend',
                           p_atts=p_atts, split=False, strategy=1)

        elif strategy == 3:
            query = "|".join(queries)
            cqp.Exec(f'{name}={query};')
            df = self.matches(cqp, name, p_atts=p_atts,
                              split=False, flags=None, strategy=2)

        # post-process dataframe
        df["freq"] = df["freq"].astype(int)
        # df = df.loc[df["freq"] != 0]

        return df


def score_counts(df, order='log_likelihood', cut_off=1000,
                 flags=None, ams=None, digits=6, vocab=None):
    """score counts in DataFrame.

    :param DataFrame df: DataFrame with reasonably-named columns, index 'item'

    :param str order: association measure for sorting (in descending order)
    :param int cut_off: number of items to retrieve
    :param str flags: '%c' / '%d' / '%cd' (cwb-ccc algorithm)
    :param list ams: association measures to calculate (None=all)
    :param int digits: round dataframe

    :return: scored counts
    :rtype: ScoreFrame

    """

    logger.info(f'scoring {len(df)} counts')

    # post-processing: fold items
    df = fold_df(df, flags)

    # calculate associations
    df = measures.score(df, measures=ams, freq=True, per_million=True,
                        digits=digits, boundary='poisson', vocab=vocab)

    # sort
    df = df.sort_values(by=[order, 'item'], ascending=[False, True])

    # apply cut-off
    df = df.head(cut_off) if cut_off is not None else df

    return df



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""collocates.py

definition of dump2cooc() and the Collocates class.

"""
import logging
from itertools import chain

# requirements
from pandas import DataFrame, to_numeric

# part of module
from .cache import generate_idx
from .counts import score_counts
from .utils import node2cotext

logger = logging.getLogger(__name__)


class Collocates:
    """ collocation analysis """

    def __init__(self, corpus, df_dump, p_query=['lemma'], mws=10,
                 df_cooc=None, f1_set=None, node_freq=None):

        # consistency check
        if df_dump is not None and len(df_dump) == 0:
            self.f1_set = set()
            logger.warning('empty dump')
            return

        # init corpus
        self.corpus = corpus

        # maximum window size (= context)
        self.mws = mws

        # determine layer to work on
        self.p_query = [p_query] if isinstance(p_query, str) else p_query
        available_attributes = self.corpus.available_attributes()['attribute'].values
        if not set(self.p_query).issubset(set(available_attributes)):
            logger.warning('specified p-attribute(s) ("%s") not available' % " ".join(self.p_query))
            logger.warning('falling back to primary layer')
            self.p_query = ['word']

        # get df_cooc, f1_set, node_freq
        if df_dump is not None:

            # get from cache if possible
            identifier = generate_idx(
                [df_dump.reset_index()[['match', 'matchend', 'context', 'contextend']], mws, p_query]
            )
            f1_set = self.corpus.cache.get(identifier + "-f1_set")
            df_cooc = self.corpus.cache.get(identifier + "-df_cooc")
            node_freq = self.corpus.cache.get(identifier + "-node_freq")

            # create and cache otherwise
            if not isinstance(df_cooc, DataFrame):
                # create collocation database
                logger.info('creating')
                logger.info('.. collecting cpos of matches and context')
                df_cooc, f1_set = dump2cooc(df_dump, self.mws)
                node_freq = self.corpus.counts.cpos(f1_set, self.p_query)
                logger.info(f'.. collected {len(df_cooc)} corpus positions')
                self.corpus.cache.set(identifier + "-f1_set", f1_set)
                self.corpus.cache.set(identifier + "-df_cooc", df_cooc)
                self.corpus.cache.set(identifier + "-node_freq", node_freq)

        elif df_cooc is None or f1_set is None or node_freq is None:
            logger.error('if no dump is given, you have to provide all frequencies')
            return

        self.identifier = generate_idx(
            [df_cooc.head(min(100, len(df_cooc))).reset_index(), f1_set, node_freq]
        )
        self.df_cooc = df_cooc
        self.f1_set = f1_set
        self.node_freq = node_freq

    def count(self, window):

        # check window
        mws = window if self.mws is None else self.mws
        if window > mws:
            logger.warning('requested window outside maximum window size')
            window = mws

        # slice window
        logger.info(f'slicing window {window}')
        relevant = self.df_cooc.loc[abs(self.df_cooc['offset']) <= window]

        # frequency counts
        f = self.corpus.counts.cpos(relevant['cpos'], self.p_query)

        return f

    def show(self, window=5, order='log_likelihood', cut_off=100,
             ams=None, min_freq=2, flags=None, marginals='corpus',
             show_negative=False):

        # consistency check
        if len(self.f1_set) == 0:
            logger.error("nothing to show")
            return DataFrame()

        identifier = generate_idx(
            [self.identifier, window, ams, min_freq, marginals, flags]
        )

        collocates = self.corpus.cache.get(identifier)

        if not isinstance(collocates, DataFrame):

            logger.info('.. counting')
            # get window counts and apply min freq
            f = self.count(window).rename(columns={'freq': 'f'})
            vocab = len(f)
            f1 = f['f'].sum()
            f = f.loc[f['f'] >= min_freq]

            # get reference frequencies
            logger.info('.. reference')
            if isinstance(marginals, str):
                if marginals == 'corpus':
                    N = self.corpus.corpus_size - len(self.f1_set)
                    marginals = self.corpus.marginals(f.index, self.p_query)
                else:
                    raise NotImplementedError
            elif isinstance(marginals, DataFrame):
                # DataFrame must contain a column 'freq'
                N = marginals['freq'].sum()
            else:
                raise NotImplementedError

            # f2 = marginals - node frequencies
            f2 = marginals[['freq']].rename(columns={'freq': 'marginal'}).join(
                self.node_freq[['freq']].rename(columns={'freq': 'in_nodes'})
            )
            f2['in_nodes'] = to_numeric(f2['in_nodes'], downcast='integer')
            f2 = f2.fillna(0)  # , downcast='infer')
            f2['f2'] = f2['marginal'] - f2['in_nodes']
            f2['f2'] = to_numeric(f2['f2'], downcast='integer')

            logger.info('.. calculating')
            # create dataframe
            df = f[['f']].join(f2[['f2']], how='left')
            df['f1'] = f1
            df['N'] = N
            df = df.fillna(0)  # , downcast='infer')

            # score
            collocates = score_counts(df, order=order, cut_off=None,
                                      flags=flags, ams=ams, vocab=vocab)

            # add node and marginal frequencies
            collocates = collocates.join(f2[['in_nodes', 'marginal']], how='left')

            # put in cache
            self.corpus.cache.set(identifier, collocates)

        # throw away anti-collocates by default
        if not show_negative:
            collocates = collocates.loc[collocates['O11'] >= collocates['E11']]

        # sort
        collocates = collocates.sort_values(by=[order, 'item'], ascending=[False, True])

        # apply cut-off
        collocates = collocates.head(cut_off) if cut_off is not None else collocates

        return collocates


def dump2cooc(df_dump, context=None, rm_nodes=True, drop_duplicates=True):
    """ converts df_dump to df_cooc + f1_set

    strategy:
    (1a) create overlapping contexts with nodes
    (1b) concatenate local contexts
    (2a) sort by abs(offset)
    (2b) deduplicate by cpos, keep first occurrences (=smallest offset)
    (3a) f1_set = (cpos where offset == 0)
    (3b) remove rows where cpos in f1_set

    :param DataFrame df_dump: [match, matchend] + [context_id, context, contextend]
    :param int context: confine context?

    :return: deduplicated df_cooc [match, cpos, offset] + f1_set (cpos of nodes)
    :rtype: tuple(DataFrame, set)
    """

    if context == 0:
        logger.warning("can't work on 0 context")
        return DataFrame(), set()

    # reset the index to be able to work with it
    df = df_dump.reset_index()

    if context is None:
        if (
            df['match'].values == df['context'].values
        ).all() and (
            df['matchend'].values == df['contextend'].values
        ).all():
            return DataFrame(), set()
        else:
            df['start'] = df['context']
            df['end'] = df['contextend']

    else:
        logger.info("re-confine regions by given context")
        df['start'] = df['match'] - context
        df['start'] = df[['start', 'context']].max(axis=1)
        df['end'] = df['matchend'] + context
        df['end'] = df[['end', 'contextend']].min(axis=1)

    logger.info("(1a) create local contexts")
    df = DataFrame.from_records(
        node2cotext(df['match'], df['matchend'], df['context'], df['contextend'])
    )

    logger.info("(1b) concatenate local contexts")
    df_infl = DataFrame({
        'match': list(chain.from_iterable(df['match_list'].values)),
        'cpos': list(chain.from_iterable(df['cpos_list'].values)),
        'offset': list(chain.from_iterable(df['offset_list'].values))
    })

    if drop_duplicates:
        logger.info("(2a) sort by absolute offset")
        df_infl['abs_offset'] = df_infl.offset.abs()
        df_infl = df_infl.sort_values(by=['abs_offset', 'cpos'])
        df_infl = df_infl.drop(["abs_offset"], axis=1)

        logger.info("(2b) drop duplicates")
        df_defl = df_infl.drop_duplicates(subset='cpos')
    else:
        df_defl = df_infl

    if not rm_nodes:
        return df_defl

    logger.info("(3a) identify nodes ...")
    f1_set = set(df_defl.loc[df_defl['offset'] == 0]['cpos'])

    logger.info("(3b) ... and remove them")
    df_defl = df_defl[df_defl['offset'] != 0]

    return df_defl, f1_set



import pytest

from ccc import Corpus
from ccc.cqpy import cqpy_dumps, cqpy_load, cqpy_loads, run_query

from .conftest import DATA_PATH


def get_corpus(corpus_settings, data_dir=DATA_PATH):

    return Corpus(
        corpus_settings['corpus_name'],
        registry_dir=corpus_settings['registry_dir'],
        lib_dir=corpus_settings.get('lib_dir', None),
        data_dir=data_dir
    )


def test_cqpy_load(query_files):
    query = cqpy_load(query_files['as_a_x_i_y_knowledge'])
    assert 'meta' in query
    assert 'cqp' in query
    assert query['anchors']['corrections'][1] == -1
    assert query['query']['context'] is None


def test_cqpy_dump(query_files):
    query_load = cqpy_load(query_files['as_a_x_i_y_knowledge'])
    query_reload = cqpy_loads(cqpy_dumps(query_load))
    assert query_load == query_reload


def test_run_from_cqpy(germaparl, query_files):

    corpus = get_corpus(germaparl)
    query = cqpy_load(query_files['jemand_sagt'])
    lines = run_query(corpus, query)
    assert all([v in lines.columns for v in ['word', 'entity_lemma', 'vp_lemma']])


def test_run_from_cqpy_sloppy(germaparl, query_files):

    corpus = get_corpus(germaparl)
    query = cqpy_load(query_files['jemand_sagt_sloppy'])
    lines = run_query(corpus, query)
    assert lines[['five_word']].value_counts()[''] == 440
    assert lines[['five_word']].value_counts()['nichts'] == 3
    assert lines[['entity_lemma']].value_counts()['sie'] == 273


def test_run_from_cqpy_display(germaparl, query_files):

    corpus = get_corpus(germaparl)
    query = cqpy_load(query_files['jemand_sagt_display'])
    lines = run_query(corpus, query)

    assert 'word' in lines.columns
    assert 'lemma' not in lines.columns
    assert 'entity_word' not in lines.columns
    assert 'entity_lemma' in lines.columns
    assert lines[['entity_lemma']].value_counts()['sie'] == 273
